import math
import unittest

import pytest
from numpy.testing import assert_array_equal

import riptable as rt
from riptable import *
from riptable.rt_utils import alignmk


class TestRTNumpy:
    """Tests for various functions in the rt_numpy module."""

    def test_where(self):
        arr = FA([False, True, False])
        result = where(arr, True, False)
        assert_array_equal(result, arr, err_msg=f"Results did not match for where. {arr} vs. {result}")

        x = Categorical(["a", "b", "a", "c", "b"])
        y = Categorical(["b", "a", "b", "b", "e"])
        z = FastArray(["a", "c", "f", "g", "a"])
        f = FastArray([True, False, True, False, False])
        out1 = where(f, x, y)
        out2 = where(f, x, z)
        out3 = where(f, z, y)
        assert_array_equal(out1, FastArray(["a", "a", "a", "b", "e"]))
        assert isinstance(out1, Categorical)
        assert_array_equal(out2, FastArray(["a", "c", "a", "g", "a"]))
        assert not isinstance(out2, Categorical)
        assert_array_equal(out3, FastArray(["a", "a", "f", "b", "e"]))
        assert not isinstance(out3, Categorical)

    def test_where_length_one(self):
        assert_array_equal(rt.where(rt.FA([False]))[0], np.array([]))

    def test_interp(self):
        x = interp(arange(3.0).astype(float32), [1.0, 2.0, 3.0], [1.0, 2.0, 3.0])
        y = interp(arange(3.0), [1.0, 2.0, 3.0], [1.0, 2.0, 3.0])
        z = np.interp(arange(3.0), [1.0, 2.0, 3.0], [1.0, 2.0, 3.0])
        assert_array_equal(z, y, err_msg=f"Results did not match for where. {z} vs. {y}")

    def test_unique(self):
        symb = FastArray(
            # Top SPX constituents by weight as of 20200414
            [
                b"MSFT",
                b"AAPL",
                b"AMZN",
                b"FB",
                b"JNJ",
                b"GOOG",
                b"GOOGL",
                b"BRK.B",
                b"PG",
                b"JPM",
                b"V",
                b"INTC",
                b"UNH",
                b"VZ",
                b"MA",
                b"T",
                b"HD",
                b"MRK",
                b"PFE",
                b"PEP",
                b"BAC",
                b"DIS",
                b"KO",
                b"WMT",
                b"CSCO",
            ],
        )

        x1, y1, z1 = np.unique(symb, return_index=True, return_inverse=True)
        x2, y2, z2 = unique(symb, return_index=True, return_inverse=True)
        assert_array_equal(x1, x2, err_msg=f"Results did not match for unique. {x1} vs. {x2}")
        assert_array_equal(y1, y2, err_msg=f"Results did not match for unique index. {y1} vs. {y2}")
        assert_array_equal(
            z1,
            z2,
            err_msg=f"Results did not match for unique inverse. {z1} vs. {z2}",
        )
        x2, y2, z2 = unique(Cat(symb), return_index=True, return_inverse=True)
        assert_array_equal(x1, x2, err_msg=f"Results did not match for cat unique. {x1} vs. {x2}")
        assert_array_equal(
            y1,
            y2,
            err_msg=f"Results did not match for cat unique index. {y1} vs. {y2}",
        )
        assert_array_equal(
            z1,
            z2,
            err_msg=f"Results did not match for cat unique inverse. {z1} vs. {z2}",
        )

        # round 2
        symb = FastArray(
            [
                "NIHD",
                "SPY",
                "AAPL",
                "LOCO",
                "XLE",
                "MRNA",
                "JD",
                "JD",
                "QNST",
                "RUSL",
                "USO",
                "TSLA",
                "NVDA",
                "GLD",
                "GLD",
                "ZS",
                "WDC",
                "AGEN",
                "AMRS",
                "AAPL",
                "SMH",
                "PYPL",
                "AAPL",
                "SQQQ",
                "GLD",
            ],
            unicode=True,
        )

        x1, y1, z1 = np.unique(symb, return_index=True, return_inverse=True)

        x2, y2, z2 = unique(symb, return_index=True, return_inverse=True)
        assert_array_equal(x1, x2, err_msg=f"Results did not match for unique. {x1} vs. {x2}")
        assert_array_equal(y1, y2, err_msg=f"Results did not match for unique index. {y1} vs. {y2}")
        assert_array_equal(
            z1,
            z2,
            err_msg=f"Results did not match for unique inverse. {z1} vs. {z2}",
        )

        x2, y2, z2 = unique(Cat(symb), return_index=True, return_inverse=True)
        assert_array_equal(x1, x2, err_msg=f"Results did not match for cat unique. {x1} vs. {x2}")
        assert_array_equal(
            y1,
            y2,
            err_msg=f"Results did not match for cat unique index. {y1} vs. {y2}",
        )
        assert_array_equal(
            z1,
            z2,
            err_msg=f"Results did not match for cat unique inverse. {z1} vs. {z2}",
        )

        # round 3
        x2, y2, z2 = unique(symb, return_index=True, return_inverse=True, lex=True)
        assert_array_equal(x1, x2, err_msg=f"Results did not match for cat unique. {x1} vs. {x2}")
        assert_array_equal(
            y1,
            y2,
            err_msg=f"Results did not match for cat unique index. {y1} vs. {y2}",
        )
        assert_array_equal(
            z1,
            z2,
            err_msg=f"Results did not match for cat unique inverse. {z1} vs. {z2}",
        )

    def test_alignmk(self):
        ds1 = rt.Dataset()
        ds1["Time"] = [0, 1, 4, 6, 8, 9, 11, 16, 19, 30]
        ds1["Px"] = [10, 12, 15, 11, 10, 9, 13, 7, 9, 10]

        ds2 = rt.Dataset()
        ds2["Time"] = [0, 0, 5, 7, 8, 10, 12, 15, 17, 20]
        ds2["Vols"] = [20, 21, 22, 23, 24, 25, 26, 27, 28, 29]

        # Categorical keys
        ds1["Ticker"] = rt.Categorical(["Test"] * 10)
        ds2["Ticker"] = rt.Categorical(["Test", "Blah"] * 5)
        res = alignmk(ds1.Ticker, ds2.Ticker, ds1.Time, ds2.Time)
        target = rt.FastArray([0, 0, 0, 2, 4, 4, 4, 6, 8, 8])
        assert_array_equal(res, target)

        # char array keys
        ds1["Ticker"] = rt.FastArray(["Test"] * 10)
        ds2["Ticker"] = rt.FastArray(["Test", "Blah"] * 5)
        res = alignmk(ds1.Ticker, ds2.Ticker, ds1.Time, ds2.Time)
        target = rt.FastArray([0, 0, 0, 2, 4, 4, 4, 6, 8, 8])
        assert_array_equal(res, target)

    def test_sample(self):
        # Test Dataset.sample
        ds = rt.Dataset({"num": [1, 2, 3, 4, 5], "str": ["ab", "bc", "cd", "de", "ef"]})
        ds_sample = ds.sample(3, rt.FA([True, True, True, False, True]), seed=1)
        ds_sample_expected = rt.Dataset({"num": [1, 2, 5], "str": ["ab", "bc", "ef"]})
        assert ds_sample.keys() == ds_sample_expected.keys()
        for col_name in ds_sample_expected.keys():
            assert_array_equal(
                ds_sample_expected[col_name], ds_sample[col_name], err_msg=f"Column '{col_name}' differs."
            )

        # Test FastArray.sample
        fa = rt.FA([1, 2, 3, 4, 5])
        fa_sample = fa.sample(2, rt.FA([False, True, True, False, True]), seed=1)
        fa_sample_expected = rt.FA([2, 3])
        assert_array_equal(fa_sample_expected, fa_sample)

        # Test overflow
        fa_sample = fa.sample(10, rt.FA([False, True, False, False, True]), seed=1)
        fa_sample_expected = rt.FA([2, 5])
        assert_array_equal(fa_sample_expected, fa_sample)

        # Test no filter
        fa_sample = fa.sample(2, seed=1)
        fa_sample_expected = rt.FA([2, 3])
        assert_array_equal(fa_sample_expected, fa_sample)

        # Test fancy index
        fa_sample = fa.sample(2, rt.FA([1, 3, 4]), seed=1)
        fa_sample_expected = rt.FA([2, 4])
        assert_array_equal(fa_sample_expected, fa_sample)

    def test_hstack_returns_same_type(self):
        # Create some instances of the same derived array type.
        # TODO: Implement another version of this test using a derived array type (maybe define one just for this test)
        #       that does not define it's own 'hstack' method. Or, test what happens if we e.g. hstack something like
        #       rt.Date() and a list (that could be used to create an array of that type).
        array_type = rt.Date

        arrs = [
            array_type(
                ["20210108", "20210108", "20210115", "20210115", "20210115", "20210122", "20210129", "20210129"]
            ),
            array_type(["20200102"]),
            array_type(["20190103", "20190204", "20190305"]),
        ]
        total_len = sum(map(lambda x: len(x), arrs))

        # rt.hstack() should return an instance of the same type.
        result = rt.hstack(arrs)
        assert type(result) == array_type
        assert len(result) == total_len


class TestHStackAny:
    """Tests for the rt.hstack_any (a.k.a. rt.stack_rows) function."""

    _fa1 = rt.FastArray([100, 200])
    _fa2 = rt.FastArray([111, 222])
    _dtn1 = rt.DateTimeNano("2021-10-12 01:02:03", from_tz="UTC")
    _dtn2 = rt.DateTimeNano("1980-03-04 13:14:15", from_tz="UTC")
    _ts1 = _dtn1 - _dtn2
    _ts2 = _dtn2 - _dtn1
    _ds1 = rt.Dataset({"a": 11})
    _ds2 = rt.Dataset({"b": 22})
    _pds1 = rt.PDataset(_ds1)
    _pds2 = rt.PDataset(_ds2)

    @pytest.mark.parametrize(
        "inputs,expected",
        [
            pytest.param([_fa1, _fa2], rt.FastArray, id="FastArray,FastArray"),
            pytest.param([_dtn1, _dtn2], rt.DateTimeNano, id="DateTimeNano,DateTimeNano"),
            pytest.param([_dtn1, _dtn2], rt.DateTimeNano, id="DateTimeNano,DateTimeNano"),
            pytest.param([_ts1, _ts2], rt.TimeSpan, id="TimeSpan,TimeSpan"),
            pytest.param([_ds1, _ds2], rt.Dataset, id="Dataset,Dataset"),
            pytest.param([_pds1, _pds2], None, id="PDataset,PDataset"),  # notyet
            pytest.param([_dtn1, _ts2], None, id="DateTimeNano,TimeSpan"),  # neither is base
            pytest.param([_fa1, _dtn2], rt.FastArray, id="FastArray,DateTimeNano"),
            pytest.param([_ts1, _fa2], rt.FastArray, id="TimeSpan,FastArray"),
            pytest.param([_ds1, _pds2], rt.Dataset, id="Dataset,PDataset"),
            pytest.param([_pds1, _ds2], rt.Dataset, id="PDataset,Dataset"),
            pytest.param([_fa1, _ds2], None, id="FastArray,Dataset"),
        ],
    )
    def test_hstack_any(self, inputs, expected):
        if expected is None:
            with pytest.raises(Exception):
                rt.hstack_any(inputs)
        else:
            result = rt.hstack_any(inputs)
            assert type(result) == expected


class TestMaximum:
    """Tests for the rt.maximum function."""

    def test_maximumscalar(self):
        # scalar vs. scalar
        x = maximum(3, 4.0, dtype=np.float32)
        assert isinstance(x, np.float32)
        assert x == 4.0

        # scalar vs. array
        scalar_value = 3
        x = maximum(scalar_value, arange(100), dtype=np.float32)
        assert x.dtype == np.float32
        expected = np.arange(100, dtype=np.float32)
        expected[:scalar_value] = scalar_value

        # array vs. scalar
        scalar_value = 3
        x = maximum(arange(100), scalar_value, dtype=np.float32)
        assert x.dtype == np.float32
        expected = np.arange(100, dtype=np.float32)
        expected[:scalar_value] = scalar_value


class TestMinimum:
    """Tests for the rt.minimum function."""

    def test_minimumscalar(self):
        # scalar vs. scalar
        x = minimum(3, 4.0, dtype=np.float32)
        assert isinstance(x, np.float32)
        assert x == 3.0

        # scalar vs. array
        scalar_value = 3
        x = minimum(scalar_value, arange(100), dtype=np.float32)
        assert x.dtype == np.float32
        expected = np.arange(100, dtype=np.float32)
        expected[scalar_value:] = scalar_value

        # array vs. scalar
        scalar_value = 3
        x = minimum(arange(100), scalar_value, dtype=np.float32)
        assert x.dtype == np.float32
        expected = np.arange(100, dtype=np.float32)
        expected[scalar_value:] = scalar_value


# TODO: Extend the tests in the TestNanmax / TestNanmin classes below to cover the following cases:
#   * non-array inputs (e.g. a list or set or scalar)
#   * other FastArray subclass, e.g. Date
#   * other unordered FastArray subclass, e.g. DateSpan (or at least, a subclass which doesn't
#       have a normal 3-way comparison because it's not a poset).
#       * For example, rt.DateSpan should probably use Allen's interval algebra which has 13 possible outcomes.
#       * Seems like we should disallow such classes; how should we check though?
#   * 2d and 3d arrays (these will probably punt to numpy, but verify)
#       * this won't work as expected on an integer dtype array; need to detect this case and just
#         disallow it rather than returning a bad result.
#   * strided/sliced arrays
#       * for sliced arrays, we're looking for issues related to data/pointer alignment in the riptide_cpp code;
#         use a reasonable-sized array (say, 1000 elements), and check we can call the function after slicing off
#         an element from the front of the array.
#   * test the above with float, integer, and string dtypes; make sure riptable invalid values (for integers; strings too?) are respected.
#   * check that any kwargs are validated (e.g. calling with a 1D array and axis=1 is not allowed).


class TestNanmax:
    """Tests for the rt.nanmax function."""

    @pytest.mark.parametrize(
        "arg",
        [
            pytest.param(math.nan, id="scalar"),
            pytest.param([math.nan, math.nan, math.nan, math.nan, math.nan], id="list-float"),
            pytest.param({np.nan}, id="set-float", marks=pytest.mark.skip("Broken in both numpy and riptable.")),
            pytest.param(np.full(100, np.nan, dtype=np.float32), id="ndarray-float"),
            pytest.param(rt.full(100, np.nan, dtype=np.float32), id="FastArray-float")
            # TODO: Ordered Categorical with all invalids
        ],
    )
    def test_allnans(self, arg):
        # Call rt.nanmax with the test input.
        # It should raise a RuntimeWarning when given an input which
        # has all NaNs **on the specified axis**.
        with pytest.warns(RuntimeWarning):
            result = rt.nanmax(arg)

        # If given a scalar or 1D array (or some collection converted to such)
        # the result should be a NaN; for higher-rank arrays, the result should
        # be an array where one of the dimensions was collapsed and if there were
        # all NaNs along the selected axis there'll be a NaN there in the result.
        # TODO: Need to fix this to assert correctly for when rt.nanmax called with a higher-rank array.
        assert rt.isnan(result)

    @pytest.mark.parametrize(
        "arg",
        [
            pytest.param([], id="list-float"),
            pytest.param({}, id="set-float", marks=pytest.mark.skip("Broken in both numpy and riptable.")),
            pytest.param(np.array([], dtype=np.float32), id="ndarray-float"),
            pytest.param(
                rt.FastArray([], dtype=np.float32),
                id="FastArray-float",
                marks=pytest.mark.xfail(
                    reason="RIP-417: The call to riptide_cpp via the ledger returns None, which then causes the isnan() to raise a TypeError. This needs to be fixed so we raise an error like numpy (either by checking for this and raising the exception, or fixing the way the function punts to numpy."
                ),
            )
            # TODO: Empty ordered Categorical (create with some categories but an empty backing array).
            # TODO: Empty Date array (representing a FastArray subclass)
        ],
    )
    def test_empty(self, arg):
        # Call rt.nanmax with an empty input -- it should raise a ValueError.
        with pytest.raises(ValueError):
            rt.nanmax(arg)

    def test_unordered_categorical_disallowed(self):
        """Test which verifies rt.nanmax raises an exception if called with an unordered Categorical."""
        # Create an unordered Categorical.
        cat = rt.Categorical(["PA", "NY", "NY", "AL", "LA", "PA", "CA", "IL", "IL", "FL", "FL", "LA"], ordered=False)
        assert not cat.ordered

        with pytest.raises(ValueError):
            rt.nanmax(cat)

    @pytest.mark.xfail(reason="RIP-417: nanmax does not (yet) support returning a category scalar.")
    def test_ordered_categorical_returns_scalar(self):
        """
        Test which verifies rt.nanmax returns a scalar (Python object or numpy scalar) representing the max Category given an ordered Categorical.
        """
        # Create an ordered Categorical (aka 'ordinal').
        cat = rt.Categorical(
            ["PA", "NY", "", "NY", "AL", "LA", "PA", "", "CA", "IL", "IL", "FL", "FL", "LA"], ordered=True
        )
        assert cat.ordered

        result = rt.nanmax(cat)

        # The result should either be a Python string, a numpy string scalar, or a Categorical scalar (if we implement one).
        is_py_str = isinstance(result, (bytes, str))
        is_np_scalar = isinstance(result, (np.bytes_, np.str_))
        is_rt_cat = isinstance(result, rt.Categorical)
        assert is_py_str or is_np_scalar or is_rt_cat

        # Check the result is correct.
        assert result == "FL"


class TestNanmin:
    """Tests for the rt.nanmin function."""

    @pytest.mark.parametrize(
        "arg",
        [
            pytest.param(math.nan, id="scalar"),
            pytest.param([math.nan, math.nan, math.nan, math.nan, math.nan], id="list-float"),
            pytest.param({np.nan}, id="set-float", marks=pytest.mark.skip("Broken in both numpy and riptable.")),
            pytest.param(np.full(100, np.nan, dtype=np.float32), id="ndarray-float"),
            pytest.param(rt.full(100, np.nan, dtype=np.float32), id="FastArray-float")
            # TODO: Ordered Categorical with all invalids
        ],
    )
    def test_allnans(self, arg):
        # Call rt.nanmin with the test input.
        # It should raise a RuntimeWarning when given an input which
        # has all NaNs **on the specified axis**.
        with pytest.warns(RuntimeWarning):
            result = rt.nanmin(arg)

        # If given a scalar or 1D array (or some collection converted to such)
        # the result should be a NaN; for higher-rank arrays, the result should
        # be an array where one of the dimensions was collapsed and if there were
        # all NaNs along the selected axis there'll be a NaN there in the result.
        # TODO: Need to fix this to assert correctly for when rt.nanmin called with a higher-rank array.
        assert rt.isnan(result)

    @pytest.mark.parametrize(
        "arg",
        [
            pytest.param([], id="list-float"),
            pytest.param({}, id="set-float", marks=pytest.mark.skip("Broken in both numpy and riptable.")),
            pytest.param(np.array([], dtype=np.float32), id="ndarray-float"),
            pytest.param(
                rt.FastArray([], dtype=np.float32),
                id="FastArray-float",
                marks=pytest.mark.xfail(
                    reason="RIP-417: The call to riptide_cpp via the ledger returns None, which then causes the isnan() to raise a TypeError. This needs to be fixed so we raise an error like numpy (either by checking for this and raising the exception, or fixing the way the function punts to numpy."
                ),
            )
            # TODO: Empty ordered Categorical (create with some categories but an empty backing array).
            # TODO: Empty Date array (representing a FastArray subclass)
        ],
    )
    def test_empty(self, arg):
        # Call rt.nanmin with an empty input -- it should raise a ValueError.
        with pytest.raises(ValueError):
            rt.nanmin(arg)

    def test_unordered_categorical_disallowed(self):
        """Test which verifies rt.nanmin raises an exception if called with an unordered Categorical."""
        # Create an unordered Categorical.
        cat = rt.Categorical(["PA", "NY", "NY", "AL", "LA", "PA", "CA", "IL", "IL", "FL", "FL", "LA"], ordered=False)
        assert not cat.ordered

        with pytest.raises(ValueError):
            rt.nanmin(cat)

    @pytest.mark.xfail(reason="RIP-417: nanmin does not (yet) support returning a category scalar.")
    def test_ordered_categorical_returns_scalar(self):
        """
        Test which verifies rt.nanmin returns a scalar (Python object or numpy scalar) representing the min Category given an ordered Categorical.
        """
        # Create an ordered Categorical (aka 'ordinal').
        cat = rt.Categorical(
            ["PA", "NY", "", "NY", "AL", "LA", "PA", "", "CA", "IL", "IL", "FL", "FL", "LA"], ordered=True
        )
        assert cat.ordered

        result = rt.nanmin(cat)

        # The result should either be a Python string, a numpy string scalar, or a Categorical scalar (if we implement one).
        is_py_str = isinstance(result, (bytes, str))
        is_np_scalar = isinstance(result, (np.bytes_, np.str_))
        is_rt_cat = isinstance(result, rt.Categorical)
        assert is_py_str or is_np_scalar or is_rt_cat

        # Check the result is correct.
        assert result == "PA"


if __name__ == "__main__":
    tester = unittest.main()



import gc
import sys
import warnings
from collections import deque
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional, Set, Tuple

import numpy as np
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
)

import riptable as rt
from riptable import Categorical
from riptable.tests.test_utils import verbose_categorical

if TYPE_CHECKING:
    from ..rt_categorical import Categorical

name = lambda obj: obj.__class__.__name__


def assert_equal_(actual, expected, decimal=7, err_msg="", verbose=True):
    """
    A wrapper around numpy testing helpers for assertions that defer to an exact or approximate assertion
    depending on the type.

    Parameters
    ----------
    actual
        Object to check.
    expected
        Desired object.
    decimal
        Desired decimal precision (default 7).
    err_msg
        Error message in event of failure (default '').
    verbose
        In event of failure with verbose set to True, the actual and expected are part of the error message (default True).

    Raises
    ------
    AssertionError
        If actual not equal to expected up to specified precision depending if the expected type is inexact.

    See Also
    --------
    assert_equal, assert_almost_equal

    Note
    ----
    Default params resemble numpy assert_equal* default params.
    """
    # TODO delegate to corresponding assertion based on type of expected object; e.g.,
    # int types would defer to assert_equal, while float types would defer to assert_almost_equal
    # Need to consider objects and other types as opposed to just the numeric hierarchy.
    try:
        assert_equal(actual, expected, err_msg, verbose)
    except AssertionError:
        assert_almost_equal(actual, expected, decimal, err_msg, verbose)


def assert_array_equal_(actual, expected, decimal=6, err_msg="", verbose=True):
    """
    A wrapper around numpy testing helpers for array assertion that are aware of dtypes and dispatch
    to exact or approximate equality assertions.

    Parameters
    ----------
    actual: array_like
        The actual array-like object to check.
    expected: array_like
        The expected array-like object.
    decimal
        Desired decimal precision (default 7).
    err_msg
        Error message in event of failure (default '').
    verbose
        In event of failure with verbose set to True, the actual and expected are part of the error message (default True).

    Raises
    ------
    AssertionError
        If actual not equal to expected up to specified precision depending if the expected type is inexact.   ------

    See Also
    --------
    assert_array_equal, assert_allclose, assert_array_almost_equal_nulp, assert_array_max_ulp, assert_equal

    Note
    ----
    Default params resemble numpy assert_array_equal* default params.
    """
    # TODO delegate to corresponding assertion based on dtype
    # E.g., exact equality for integer types and approximate equality with tolerance for floating types
    # Need to consider objects and other types as opposed to just the numeric hierarchy.
    try:
        assert_array_equal(np.array(actual), expected)
    except AssertionError:
        assert_allclose(np.array(actual), expected)


def assert_categorical_equal(
    actual: Categorical,
    expected: Categorical,
    check_expanded_array: Optional[bool] = True,
    verbose: Optional[bool] = False,
) -> None:
    """
    Assert ``actual`` and ``expected`` Categoricals, of the various ``CategoryModes``, have no mismatch.

    Parameters
    ----------
    actual: Categorical
    expected: Categorical
    check_expanded_array: bool, optional
        Check the expanded Categoricals are equal (default True).
    verbose: bool, optional
        Logs a verbose representation of ``actual`` and ``expected`` Categorical on failure.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If ```actual`` or ``expected`` are not of type ``Categorical``.
    AssertionError
        If any mismatch in ``actual`` and ``expected`` exists.

    Note
    ----
    ``check_expanded_array`` is expensive since it will re-construct the array the Categorical
    is derived from. It may be sufficient to check the underlying ``FastArray`` as well as
    the categories.
    """
    fn = "assert_categorical_equal"
    try:
        # check actual and expected types
        if not isinstance(expected, Categorical):
            raise ValueError(f'{fn}: expected type Categorical, got {type(expected)} for "expected" parameter')
        if not isinstance(actual, Categorical):
            raise ValueError(f'{fn}: expected type Categorical, got {type(actual)} for "actual" parameter')

        # check category_mode
        assert (
            actual.category_mode == expected.category_mode
        ), f"{fn}: category mode mismatch\nactual {repr(rt.rt_enum.CategoryMode(actual.category_mode))}\nexpected {repr(rt.rt_enum.CategoryMode(expected.category_mode))}"

        # check category_array
        if (
            expected.category_mode == rt.rt_enum.CategoryMode.StringArray
            or expected.category_mode == rt.rt_enum.CategoryMode.NumericArray
        ):
            assert_array_equal(
                actual.category_array, expected.category_array, err_msg=f'{fn}: mismatch in "category_array"'
            )
            assert_array_equal(
                actual.category_array, expected.category_array, err_msg=f'{fn}: mismatch in "category_array"'
            )
        else:  # TODO implement category_mode checks for MultiKey, Dict, and IntEnum.
            warnings.warn(
                f"{fn}: category_array checks not implemented for {repr(rt.rt_enum.CategoryMode(expected.category_mode))} "
            )

        # check underlying FastArray
        if (
            expected.category_mode == rt.rt_enum.CategoryMode.StringArray
            or expected.category_mode == rt.rt_enum.CategoryMode.NumericArray
        ):
            assert_array_equal(actual._fa, expected._fa, err_msg=f"{fn}: mismatch in underlying FastArray")

        # check expand_array
        if check_expanded_array:
            assert_array_equal(actual.expand_array, expected.expand_array, err_msg=f'{fn}: mismatch in "expand_array"')

        # check category_dict
        for k, fa in expected.category_dict.items():
            assert_array_equal(
                actual.category_dict.get(k, None),
                fa,
                err_msg=f'{fn}: mismatch "category_dict" for item "{k}"\nactual keys {actual.category_dict.keys()}\nexpected keys {expected.category_dict.keys()}',
            )

        # TODO check CategoryMode Dictionary category_mapping and category_codes

    except Exception:
        if verbose:
            print(f"{fn}: expected\n{verbose_categorical(expected)}\n" f"{fn}: actual\n{verbose_categorical(actual)}\n")
        raise


def get_common_and_diff_members(obj_a: object, obj_b: object) -> Tuple[Set[str], Set[str]]:
    """
    Return the commonalities and differences between the two objects public API.

    Parameters
    ----------
    obj_a: object
        Object to find commonalities and difference when compared with obj_b.
    obj_b: object
        Object to find commonalities and difference when compared with obj_a.

    Returns
    -------
    tuple
        A tuple of two sets of strings where the first is the commonalities and second is the differences.
    """
    is_public = lambda name: not (name.startswith("__") or name.startswith("_"))
    obj_a_dir, obj_b_dir = (
        set(filter(is_public, dir(obj_a))),
        set(filter(is_public, dir(obj_b))),
    )
    common_members = obj_a_dir.intersection(obj_b_dir)
    diff_members = obj_a_dir.symmetric_difference(obj_b_dir)
    return (common_members, diff_members)


# 20200304 / Riptable version 1.3.367 - Below are the common and differing members of ndarray vs FastArray:
# common members: {'flags', 'fill', 'partition', 'transpose', 'repeat', 'real', 'itemsize', 'ctypes', 'getfield', 'itemset', 'newbyteorder', 'tobytes', 'take', 'dumps', 'conjugate', 'flatten', 'nonzero', 'setflags', 'searchsorted', 'mean', 'dump', 'dtype', 'copy', 'data', 'imag', 'max', 'byteswap', 'astype', 'std', 'argsort', 'any', 'item', 'resize', 'shape', 'dot', 'compress', 'tostring', 'min', 'tofile', 'base', 'flat', 'tolist', 'ndim', 'cumprod', 'T', 'reshape', 'trace', 'ravel', 'squeeze', 'argmin', 'sum', 'sort', 'prod', 'diagonal', 'nbytes', 'argmax', 'strides', 'conj', 'ptp', 'round', 'all', 'put', 'size', 'clip', 'var', 'swapaxes', 'argpartition', 'cumsum', 'choose', 'view', 'setfield'}
# differing members: {'count', 'where', 'nanvar', 'WarningDict', 'clip_lower', 'str', 'isna', 'isnotnan', 'clip_upper', 'isin', 'Verbose', 'diff', 'push', 'median', 'differs', 'rolling_nanstd', 'move_mean', 'fill_backward', 'nanrankdata', 'describe', 'fill_invalid', 'MAX_DISPLAY_LEN', 'numbastring', 'SafeConversions', 'register_function', 'timewindow_sum', 'rolling_var', 'fillna', 'isfinite', 'duplicated', 'info', 'nanmean', 'ema_decay', 'nanargmax', 'save', 'rolling_nanvar', 'FasterUFunc', 'iscomputable', 'move_rank', 'tile', 'copy_invalid', 'sample', 'apply', 'move_argmax', 'get_name', 'move_median', 'cummax', 'Recycle', 'trunc', 'issorted', 'move_max', 'rolling_std', 'nunique', 'move_std', 'fill_forward', 'isnotnormal', 'replacena', 'crc', 'apply_pandas', 'notna', 'timewindow_prod', 'nanmax', 'CompressPickle', 'NoTolerance', 'abs', 'isinf', 'move_min', 'rolling_nansum', 'isnotinf', 'set_name', 'normalize_minmax', 'nansum', 'rolling_mean', 'isnan', 'isnotfinite', 'move_sum', 'argpartition2', 'inv', 'replace', 'rankdata', 'transitions', 'doc', 'normalize_zscore', 'rolling_nanmean', 'partition2', 'str_append', 'nanstd', 'display_query_properties', 'shift', 'move_var', 'unique', 'nanargmin', 'apply_schema', 'map', 'cummin', 'move_argmin', 'isnormal', 'map_old', 'nanmin', 'rolling_sum', 'WarningLevel', 'sign', 'isnanorzero', 'apply_numba'}
_NDARRAY_FASTARRAY_COMMON_AND_DIFF = get_common_and_diff_members(np.ndarray, rt.FastArray)


def get_size(obj: object) -> int:
    """Returns the total size, in bytes, of the object and the object's referents using a
    breadth-first search sweep."""
    # TODO use the robust and well tests Pympler tool to measure the size of the objects
    # Pympler tool has a asizeof that calculates the combined size in bytes of an object.
    # Some shortcomings of this approach is that it gets all referents including of the object
    # class definition, decorators, and other referents that are not relevant to the objects
    # dynamic memory footprint.
    if not isinstance(obj, Iterable):
        obj = [obj]
    size: int = 0
    seen: Set[int] = set()
    queue: deque = deque(obj)
    while len(queue):
        u = queue.popleft()
        refs = gc.get_referents(u)
        for ref in refs:
            ref_id = id(ref)
            if ref_id not in seen:  # guard against cycles
                seen.add(ref_id)
                queue.append(ref)
        size += sys.getsizeof(u)
    return size



"""
Time/ordering-based merge implementations.
"""
__all__ = ["merge_asof2"]

from datetime import timedelta
import logging
from time import perf_counter_ns
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple, Union
import warnings

import numpy as np
import numpy.typing as npt
import numba as nb

from .config import get_global_settings
from .numba.indexing import scalar_or_lookup
from .numba.invalid_values import get_invalid, get_max_valid, get_min_valid, is_valid
from .rt_fastarray import FastArray
from .rt_grouping import Grouping
from .rt_datetime import Date, DateTimeNano, DateTimeNanoScalar, TimeSpanScalar
from .rt_numpy import empty, full, full_like, ismember, isnan, isnotnan
from .rt_merge import (
    JoinIndices,
    _construct_colname_mapping,
    _extract_on_columns,
    _gbkeys_extract,
    _get_or_create_keygroup,
    _normalize_selected_columns,
    _require_columns_present,
    _verify_join_keys_compat,
)

if TYPE_CHECKING:
    from .rt_dataset import Dataset


_logger = logging.getLogger(__name__)
"""Logger for this module."""


@nb.generated_jit(nopython=True)
def _less_than_comparison(allow_exact_matches, x, y):
    """
    Numba-based wrapper for ``__lt__`` and ``__le__`` which allows them to be
    selected at compile-time using type-based dispatch.

    Parameters
    ----------
    allow_exact_matches : bool, optional
    x, y
        Values to compare.

    Returns
    -------
    bool
        ``x < y`` if `allow_exact_matches` is ``None``; otherwise ``x <= y``.
    """
    if isinstance(allow_exact_matches, nb.types.NoneType):
        return lambda allow_exact_matches, x, y: x < y
    else:
        return lambda allow_exact_matches, x, y: x <= y


@nb.generated_jit(nopython=True)
def _greater_than_comparison(allow_exact_matches, x, y):
    """
    Numba-based wrapper for ``__gt__`` and ``__ge__`` which allows them to be
    selected at compile-time using type-based dispatch.

    Parameters
    ----------
    allow_exact_matches : bool, optional
    x, y
        Values to compare.

    Returns
    -------
    bool
        ``x > y`` if `allow_exact_matches` is ``None``; otherwise ``x >= y``.
    """
    if isinstance(allow_exact_matches, nb.types.NoneType):
        return lambda allow_exact_matches, x, y: x > y
    else:
        return lambda allow_exact_matches, x, y: x >= y


@nb.njit(cache=get_global_settings().enable_numba_cache)
def _merge_asof_backward(
    haystack_2nd_key: np.ndarray,
    needles_2nd_key: np.ndarray,
    needle_idx_in_haystack: npt.NDArray[np.integer],
    haystack_first_unsorted_idx: npt.NDArray[np.integer],
    needle_first_unsorted_idx: npt.NDArray[np.integer],
    equality_tolerance: Optional[Any],
    allow_exact_matches: Optional[Any],
) -> None:
    """
    Numba-based implementation of the 'backward'-mode 'merge_asof' without groups
    (when no 'by' column(s) are specified).

    Parameters
    ----------
    haystack_2nd_key : np.ndarray
        The 'on' column data from the 'right' `Dataset`.
    needles_2nd_key : np.ndarray
        The 'on' column data from the 'left' `Dataset`.
    needle_idx_in_haystack : np.ndarray of integer
        Output array, which, after this function returns, will contain an integer fancy index
        where each element indicates an index within `haystack` where
        The shape must be the same as `needles_igroup` and the dtype should match `haystack_igroup`.
    haystack_first_unsorted_idx, needle_first_unsorted_idx : np.ndarray of bool
        Single-element integer arrays (output) to hold indices within `haystack_2nd_key` and `needles_2nd_key`,
        respectively, where the first unsorted/out-of-order element was detected within that array (if any).
        Pre-initialize these arrays with invalid values so it's easy to detect when elements have been set.
        The arrays are checked after this function returns and are used to provide better error messages
        to users if unsorted values are detected.
    equality_tolerance : scalar, optional
    allow_exact_matches : bool, optional
        Indicates whether exact matches are allowed for the 'on' values.
        Specify ``None`` for ``False`` and ``True`` for ``True``;
        this leverages numba's type-based dispatch mechanism to specialize
        the compiled code with the appropriate comparison function.
    """
    # Get the 'invalid' value for the output array, to use when
    # we can't find a match.
    invalid_haystack_idx = get_invalid(needle_idx_in_haystack[0])

    # The lengths of the 'haystack' and 'needles' arrays.
    haystack_length = len(haystack_2nd_key)
    needles_length = len(needles_2nd_key)

    # Initialize needle_key2_curr, haystack_key2_curr
    # to minimum VALID value for their dtype (making sure not to set to riptable invalid).
    # If we try to be clever about declaring and initializing them in the loop below, numba isn't
    # able to infer the types and can't compile the function.
    # Initializing the values here -- to the same value for 'prev' and 'curr' -- also allows
    # some checks (e.g. 'if i_needle > 0') to be elided in the loop body.
    haystack_key2_curr = get_min_valid(haystack_2nd_key)
    needle_key2_curr = get_min_valid(needles_2nd_key)

    # Walk through the needle and haystack 2nd key ("on") values in a linear,
    # mergesort-style loop. This relies on both arrays (from their respective groups)
    # being sorted; if only the 'haystack' 2nd key values were sorted (within the group)
    # we'd have to implement this as a 'searchsorted'-style binary search
    # for each needle. (Which is faster for one element, but the linear style is much
    # faster for matching a whole array of elements.)
    i_needle = 0
    i_haystack = 0
    while i_needle < needles_length and i_haystack < haystack_length:
        # For convenience / conciseness in the rest of the loop.
        needle_curr_idx = i_needle
        haystack_curr_idx = i_haystack

        # Get the current 'needle' and 'haystack' elements from their 2nd keys (the 'on' arrays).
        # Check them against the previous 'needle' and 'haystack' values, respectively;
        # this in-line sortedness check on the data is more efficient (faster) than doing the
        # check separately, e.g. as a separate method.
        needle_key2_prev = needle_key2_curr
        needle_key2_curr = needles_2nd_key[needle_curr_idx]
        haystack_key2_prev = haystack_key2_curr
        haystack_key2_curr = haystack_2nd_key[haystack_curr_idx]

        # TODO: Should we also (maybe optionally?) recognize riptable integer invalid values
        #       here and consider them like NaN's w.r.t. sortedness?
        # TODO: Should we provide an option to users to ignore NaN (or any other unsorted value)?
        if not (needle_key2_prev <= needle_key2_curr):
            # Detected unsorted (and/or NaN) 'on' value in needle.
            # Save the current index of the needle (within the full array, not the group)
            # into the "unsorted index" array at this group's index,
            # then skip processing for the rest of this group because the results
            # wouldn't be meaningful anyway.
            needle_first_unsorted_idx[0] = needle_curr_idx
            i_needle = needles_length
            continue

        if not (haystack_key2_prev <= haystack_key2_curr):
            # Detected unsorted (and/or NaN) 'on' value in haystack.
            haystack_first_unsorted_idx[0] = haystack_curr_idx
            i_haystack = haystack_length
            continue

        # Compare the current values of the 2nd key for needle and haystack.
        if _less_than_comparison(allow_exact_matches, haystack_key2_curr, needle_key2_curr):
            # If we haven't reached the end of the haystack, check to see if the
            # next haystack value is *also* greater than (or equal to) the current needle.
            # If so, the *current* haystack value is the one we want the needle to match against,
            # because it's the largest haystack value that's still less than or equal to the needle.
            # Otherwise, if the next haystack is still less than the current needle,
            # advance the haystack pointer and recurse.
            if i_haystack + 1 < haystack_length and _less_than_comparison(
                allow_exact_matches, haystack_2nd_key[i_haystack + 1], needle_key2_curr
            ):
                i_haystack += 1

            else:
                # If the current value of `needles` 2nd key is greater than or equal to the current value of `haystack` 2nd key,
                # and their difference is *at most* the tolerance value (if specified), we found a match,
                # so assign the current index within the haystack array to the current index of the needles array.
                # Otherwise, the haystack value was outside the tolerance value so this isn't a match.
                needle_idx_in_haystack[needle_curr_idx] = (
                    haystack_curr_idx
                    # TODO: Do we need to use _less_than_comparison for the equality_tolerance check here?
                    #       Or does it always allow exact matches (i.e. the difference between the values is exactly
                    #       equal to the tolerance)?
                    if equality_tolerance is None or equality_tolerance >= (needle_key2_curr - haystack_key2_curr)
                    else invalid_haystack_idx
                )

                i_needle += 1

        else:
            # N.B. This assumes haystack_key2_curr > needle_key2_curr, but it might not
            # if the 2nd keys are floats and the current value is NaN.

            # This method implements the 'backward'-looking merge_asof, so if we're at
            # the start of the haystack, we know there's no match for this needle.
            if i_haystack < 1:
                needle_idx_in_haystack[needle_curr_idx] = invalid_haystack_idx
                i_needle += 1
            else:
                i_haystack += 1

    # If we reached the end of the haystack and have some remaining needles,
    # they should all be greater than (or equal to) the last/greatest haystack value.
    # Set their corresponding values in the output array to point to the last haystack value.
    if i_needle < needles_length:
        needle_idx_in_haystack[i_needle:] = haystack_length - 1 if haystack_length > 0 else invalid_haystack_idx


@nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache)
def _merge_asof_grouped_backward(
    haystack_igroup: npt.NDArray[np.integer],
    haystack_ifirstgroup: npt.NDArray[np.integer],
    haystack_ncountgroup: npt.NDArray[np.integer],
    haystack_2nd_key: np.ndarray,
    needles_igroup: npt.NDArray[np.integer],
    needles_ifirstgroup: npt.NDArray[np.integer],
    needles_ncountgroup: npt.NDArray[np.integer],
    needles_2nd_key: np.ndarray,
    needle_key_to_haystack_key: npt.NDArray[np.integer],
    needle_idx_in_haystack: npt.NDArray[np.integer],
    haystack_first_unsorted_idx_in_group: npt.NDArray[np.integer],
    needle_first_unsorted_idx_in_group: npt.NDArray[np.integer],
    equality_tolerance: Optional[Any],
    allow_exact_matches: Optional[Any],
) -> None:
    """
    Numba-based implementation of the 'backward'-mode 'merge_asof' over groups
    (when 'by' columns are specified).

    Parameters
    ----------
    haystack_igroup, haystack_ifirstgroup, haystack_ncountgroup : np.ndarray of integer
        Arrays from the `Grouping` created for the 'by' column(s) of the 'right' `Dataset`.
    haystack_2nd_key : np.ndarray
        The 'on' column data from the 'right' `Dataset`.
    needles_igroup, needles_ifirstgroup, needles_ncountgroup : np.ndarray of integer
        Arrays from the `Grouping` created for the 'by' column(s) of the 'left' `Dataset`.
    needles_2nd_key : np.ndarray
        The 'on' column data from the 'left' `Dataset`.
    needle_key_to_haystack_key : np.ndarray of integer
        Array mapping indices of 'needle' "by" keys to the index of the same key
        from 'haystack', if the key also exists there.
    needle_idx_in_haystack : np.ndarray of integer
        Output array, which, after this function returns, will contain an integer fancy index
        where each element indicates an index within `haystack` where
        The shape must be the same as `needles_igroup` and the dtype should match `haystack_igroup`.
    haystack_first_unsorted_idx_in_group, needle_first_unsorted_idx_in_group : np.ndarray of integer
        Integer arrays (output) same length as `haystack_ncountgroup` and `needles_ncountgroup`.
        to hold indices within `haystack_2nd_key` and `needles_2nd_key`, respectively, where the
        first unsorted/out-of-order element was detected within that group (if any).
        Pre-initialize these arrays with invalid values so it's easy to detect when elements have been set.
        The arrays are checked after this function returns and are used to provide better error messages
        to users if unsorted values are detected.
    equality_tolerance : scalar, optional
    allow_exact_matches : bool, optional
        Indicates whether exact matches are allowed for the 'on' values.
        Specify ``None`` for ``False`` and ``True`` for ``True``;
        this leverages numba's type-based dispatch mechanism to specialize
        the compiled code with the appropriate comparison function.
    """
    # Get the 'invalid' value for the output array, to use when
    # we can't find a match.
    invalid_haystack_idx = get_invalid(needle_idx_in_haystack[0])

    for needles_key_idx in nb.prange(len(needle_key_to_haystack_key)):  # pylint: disable=not-an-iterable
        # Add one (1) to convert from the key index to the group index,
        # because group 0 represents the null/NA group.
        needle_group_offset = needles_ifirstgroup[needles_key_idx + 1]
        needle_group_length = needles_ncountgroup[needles_key_idx + 1]

        # Does this 'needles' KEY (1st key only) match a 'haystack' KEY (1st key only)?
        # If so, what's the corresponding 'haystack' group index?
        haystack_key_idx = needle_key_to_haystack_key[needles_key_idx]
        if haystack_key_idx < 0:
            # This needle doesn't exist in 'haystack'.
            # Set the values of all elements in 'needle_idx_in_haystack' that match this group to the invalid value for the array.
            needle_idx_in_haystack[
                needles_igroup[needle_group_offset : needle_group_offset + needle_group_length]
            ] = invalid_haystack_idx

            # Continue processing the next group.
            continue

        # Found a match on the first key between the 'needles' and 'haystack'.
        haystack_group_offset = haystack_ifirstgroup[haystack_key_idx + 1]
        haystack_group_length = haystack_ncountgroup[haystack_key_idx + 1]

        # Initialize needle_key2_curr, haystack_key2_curr
        # to minimum VALID value for their dtype (making sure not to set to riptable invalid).
        # If we try to be clever about declaring and initializing them in the loop below, numba isn't
        # able to infer the types and can't compile the function.
        # Initializing the values here -- to the same value for 'prev' and 'curr' -- also allows
        # some checks (e.g. 'if i_needle > 0') to be elided in the loop body.
        haystack_key2_curr = get_min_valid(haystack_2nd_key)
        needle_key2_curr = get_min_valid(needles_2nd_key)

        # Walk through the needle and haystack 2nd key ("on") values in a linear,
        # mergesort-style loop. This relies on both arrays (from their respective groups)
        # being sorted; if only the 'haystack' 2nd key values were sorted (within the group)
        # we'd have to implement this as a 'searchsorted'-style binary search
        # for each needle. (Which is faster for one element, but the linear style is much
        # faster for matching a whole array of elements.)
        i_needle = 0
        i_haystack = 0
        while i_needle < needle_group_length and i_haystack < haystack_group_length:
            # For convenience / conciseness in the rest of the loop.
            needle_curr_idx = needles_igroup[needle_group_offset + i_needle]
            haystack_curr_idx = haystack_igroup[haystack_group_offset + i_haystack]

            # Get the current 'needle' and 'haystack' elements from their 2nd keys (the 'on' arrays).
            # Check them against the previous 'needle' and 'haystack' values, respectively;
            # this in-line sortedness check on the data is more efficient (faster) than doing the
            # check separately, e.g. as a separate method.
            needle_key2_prev = needle_key2_curr
            needle_key2_curr = needles_2nd_key[needle_curr_idx]
            haystack_key2_prev = haystack_key2_curr
            haystack_key2_curr = haystack_2nd_key[haystack_curr_idx]

            # TODO: Should we also (maybe optionally?) recognize riptable integer invalid values
            #       here and consider them like NaN's w.r.t. sortedness?
            # TODO: Should we provide an option to users to ignore NaN (or any other unsorted value)?
            if not (needle_key2_prev <= needle_key2_curr):
                # Detected unsorted (and/or NaN) 'on' value in needle.
                # Save the current index of the needle (within the full array, not the group)
                # into the "unsorted index" array at this group's index,
                # then skip processing for the rest of this group because the results
                # wouldn't be meaningful anyway.
                needle_first_unsorted_idx_in_group[needles_key_idx + 1] = needle_curr_idx
                i_needle = needle_group_length
                continue

            if not (haystack_key2_prev <= haystack_key2_curr):
                # Detected unsorted (and/or NaN) 'on' value in haystack.
                haystack_first_unsorted_idx_in_group[haystack_key_idx + 1] = haystack_curr_idx
                i_haystack = haystack_group_length
                continue

            # Compare the current values of the 2nd key for needle and haystack.
            if _less_than_comparison(allow_exact_matches, haystack_key2_curr, needle_key2_curr):
                # If we haven't reached the end of the haystack, check to see if the
                # next haystack value is *also* greater than (or equal to) the current needle.
                # If so, the *current* haystack value is the one we want the needle to match against,
                # because it's the largest haystack value that's still less than or equal to the needle.
                # Otherwise, if the next haystack is still less than the current needle,
                # advance the haystack pointer and recurse.
                if i_haystack + 1 < haystack_group_length and _less_than_comparison(
                    allow_exact_matches,
                    haystack_2nd_key[haystack_igroup[haystack_group_offset + i_haystack + 1]],
                    needle_key2_curr,
                ):
                    i_haystack += 1

                else:
                    # If the current value of `needles` 2nd key is greater than or equal to the current value of `haystack` 2nd key,
                    # and their difference is *at most* the tolerance value (if specified), we found a match,
                    # so assign the current index within the haystack array to the current index of the needles array.
                    # Otherwise, the haystack value was outside the tolerance value so this isn't a match.
                    needle_idx_in_haystack[needle_curr_idx] = (
                        haystack_curr_idx
                        # TODO: Do we need to use _less_than_comparison for the equality_tolerance check here?
                        #       Or does it always allow exact matches (i.e. the difference between the values is exactly
                        #       equal to the tolerance)?
                        if equality_tolerance is None or equality_tolerance >= (needle_key2_curr - haystack_key2_curr)
                        else invalid_haystack_idx
                    )

                    i_needle += 1

            else:
                # N.B. This assumes haystack_key2_curr > needle_key2_curr, but it might not
                # if the 2nd keys are floats and the current value is NaN.

                # This method implements the 'backward'-looking merge_asof, so if we're at
                # the start of the haystack, we know there's no match for this needle.
                if i_haystack < 1:
                    needle_idx_in_haystack[needle_curr_idx] = invalid_haystack_idx
                    i_needle += 1
                else:
                    i_haystack += 1

        # If we reached the end of the haystack and have some remaining needles,
        # they should all be greater than (or equal to) the last/greatest haystack value.
        # Set their corresponding values in the output array to point to the last haystack value.
        if i_needle < needle_group_length:
            needle_idx_in_haystack[
                needles_igroup[needle_group_offset + i_needle : needle_group_offset + needle_group_length]
            ] = (haystack_group_length - 1 if haystack_group_length > 0 else invalid_haystack_idx)


@nb.njit(cache=get_global_settings().enable_numba_cache)
def _merge_asof_forward(
    haystack_2nd_key: np.ndarray,
    needles_2nd_key: np.ndarray,
    needle_idx_in_haystack: npt.NDArray[np.integer],
    haystack_first_unsorted_idx: npt.NDArray[np.integer],
    needle_first_unsorted_idx: npt.NDArray[np.integer],
    equality_tolerance: Optional[Any],
    allow_exact_matches: Optional[Any],
) -> None:
    """
    Numba-based implementation of the 'forward'-mode 'merge_asof' without groups
    (when no 'by' column(s) are specified).

    Parameters
    ----------
    haystack_2nd_key : np.ndarray
        The 'on' column data from the 'right' `Dataset`.
    needles_2nd_key : np.ndarray
        The 'on' column data from the 'left' `Dataset`.
    needle_idx_in_haystack : np.ndarray of integer
        Output array, which, after this function returns, will contain an integer fancy index
        where each element indicates an index within `haystack` where
        The shape must be the same as `needles_igroup` and the dtype should match `haystack_igroup`.
    haystack_first_unsorted_idx, needle_first_unsorted_idx : np.ndarray of integer
        Single-element integer arrays (output) to hold indices within `haystack_2nd_key` and `needles_2nd_key`,
        respectively, where the first unsorted/out-of-order element was detected within that group (if any).
        Pre-initialize these arrays with invalid values so it's easy to detect when elements have been set.
        The arrays are checked after this function returns and are used to provide better error messages
        to users if unsorted values are detected.
    equality_tolerance : scalar, optional
    allow_exact_matches : bool, optional
        Indicates whether exact matches are allowed for the 'on' values.
        Specify ``None`` for ``False`` and ``True`` for ``True``;
        this leverages numba's type-based dispatch mechanism to specialize
        the compiled code with the appropriate comparison function.
    """
    # Get the 'invalid' value for the output array, to use when
    # we can't find a match.
    invalid_haystack_idx = get_invalid(needle_idx_in_haystack[0])

    # The lengths of the 'haystack' and 'needles' arrays.
    haystack_length = len(haystack_2nd_key)
    needles_length = len(needles_2nd_key)

    # Initialize needle_key2_curr, haystack_key2_curr
    # to maximum VALID value for their dtype (making sure not to set to riptable invalid).
    # If we try to be clever about declaring and initializing them in the loop below, numba isn't
    # able to infer the types and can't compile the function.
    # Initializing the values here -- to the same value for 'prev' and 'curr' -- also allows
    # some checks (e.g. 'if i_needle > 0') to be elided in the loop body.
    haystack_key2_curr = get_max_valid(haystack_2nd_key)
    needle_key2_curr = get_max_valid(needles_2nd_key)

    # Walk through the needle and haystack 2nd key ("on") values in a linear,
    # mergesort-style loop. This relies on both arrays (from their respective groups)
    # being sorted; if only the 'haystack' 2nd key values were sorted (within the group)
    # we'd have to implement this as a 'searchsorted'-style binary search
    # for each needle. (Which is faster for one element, but the linear style is much
    # faster for matching a whole array of elements.)
    i_needle = needles_length - 1
    i_haystack = haystack_length - 1
    while i_needle >= 0 and i_haystack >= 0:
        # For convenience / conciseness in the rest of the loop.
        needle_curr_idx = i_needle
        haystack_curr_idx = i_haystack

        # Get the current 'needle' and 'haystack' elements from their 2nd keys (the 'on' arrays).
        # Check them against the previous 'needle' and 'haystack' values, respectively;
        # this in-line sortedness check on the data is more efficient (faster) than doing the
        # check separately, e.g. as a separate method.
        needle_key2_prev = needle_key2_curr
        needle_key2_curr = needles_2nd_key[needle_curr_idx]
        haystack_key2_prev = haystack_key2_curr
        haystack_key2_curr = haystack_2nd_key[haystack_curr_idx]

        # TODO: Should we also (maybe optionally?) recognize riptable integer invalid values
        #       here and consider them like NaN's w.r.t. sortedness?
        # TODO: Should we provide an option to users to ignore NaN (or any other unsorted value)?
        if not (needle_key2_prev >= needle_key2_curr):
            # Detected unsorted (and/or NaN) 'on' value in needle.
            # Save the current index of the needle (within the full array, not the group)
            # into the "unsorted index" array at this group's index,
            # then skip processing for the rest of this group because the results
            # wouldn't be meaningful anyway.
            needle_first_unsorted_idx[0] = needle_curr_idx
            i_needle = -1
            continue

        if not (haystack_key2_prev >= haystack_key2_curr):
            # Detected unsorted (and/or NaN) 'on' value in haystack.
            haystack_first_unsorted_idx[0] = haystack_curr_idx
            i_haystack = -1
            continue

        # Compare the current values of the 2nd key for needle and haystack.
        if _greater_than_comparison(allow_exact_matches, haystack_key2_curr, needle_key2_curr):
            # If we haven't reached the front of the haystack, check to see if the
            # previous haystack value is *also* less than (or equal to) the current needle.
            # If so, the *current* haystack value is the one we want the needle to match against,
            # because it's the smallest haystack value that's still greater than or equal to the needle.
            # Otherwise, if the previous haystack is still greater than the current needle,
            # decrement the haystack pointer and recurse.
            if i_haystack > 0 and _greater_than_comparison(
                allow_exact_matches, haystack_2nd_key[i_haystack - 1], needle_key2_curr
            ):
                i_haystack -= 1

            else:
                # If the current value of `needles` 2nd key is greater than or equal to the current value of `haystack` 2nd key,
                # and their difference is *at most* the tolerance value (if specified), we found a match,
                # so assign the current index within the haystack array to the current index of the needles array.
                # Otherwise, the haystack value was outside the tolerance value so this isn't a match.
                needle_idx_in_haystack[needle_curr_idx] = (
                    haystack_curr_idx
                    # TODO: Do we need to use _exact_than_comparison for the equality_tolerance check here?
                    #       Or does it always allow exact matches (i.e. the difference between the values is exactly
                    #       equal to the tolerance)?
                    if equality_tolerance is None or equality_tolerance >= (haystack_key2_curr - needle_key2_curr)
                    else invalid_haystack_idx
                )

                i_needle -= 1

        else:
            # N.B. This assumes haystack_key2_curr < needle_key2_curr, but it might not
            # if the 2nd keys are floats and the current value is NaN.

            # This method implements the 'forward'-looking merge_asof, so if we're still at
            # the end/back (greatest index) of the haystack, we know there's no match for this needle.
            if i_haystack + 1 == haystack_length:
                needle_idx_in_haystack[needle_curr_idx] = invalid_haystack_idx
                i_needle -= 1
            else:
                i_haystack -= 1

    # If we reached the end of the haystack and have some remaining needles,
    # they should all be greater than (or equal to) the last/greatest haystack value.
    # Set their corresponding values in the output array to point to the last haystack value.
    if i_needle >= 0:
        needle_idx_in_haystack[:i_needle] = 0 if haystack_length > 0 else invalid_haystack_idx


@nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache)
def _merge_asof_grouped_forward(
    haystack_igroup: npt.NDArray[np.integer],
    haystack_ifirstgroup: npt.NDArray[np.integer],
    haystack_ncountgroup: npt.NDArray[np.integer],
    haystack_2nd_key: np.ndarray,
    needles_igroup: npt.NDArray[np.integer],
    needles_ifirstgroup: npt.NDArray[np.integer],
    needles_ncountgroup: npt.NDArray[np.integer],
    needles_2nd_key: np.ndarray,
    needle_key_to_haystack_key: npt.NDArray[np.integer],
    needle_idx_in_haystack: npt.NDArray[np.integer],
    haystack_first_unsorted_idx_in_group: npt.NDArray[np.integer],
    needle_first_unsorted_idx_in_group: npt.NDArray[np.integer],
    equality_tolerance: Optional[Any],
    allow_exact_matches: Optional[Any],
) -> None:
    """
    Numba-based implementation of the 'forward'-mode 'merge_asof' over groups
    (when 'by' columns are specified).

    Parameters
    ----------
    haystack_igroup, haystack_ifirstgroup, haystack_ncountgroup : np.ndarray of integer
        Arrays from the `Grouping` created for the 'by' column(s) of the 'right' `Dataset`.
    haystack_2nd_key : np.ndarray
        The 'on' column data from the 'right' `Dataset`.
    needles_igroup, needles_ifirstgroup, needles_ncountgroup : np.ndarray of integer
        Arrays from the `Grouping` created for the 'by' column(s) of the 'left' `Dataset`.
    needles_2nd_key : np.ndarray
        The 'on' column data from the 'left' `Dataset`.
    needle_key_to_haystack_key : np.ndarray of integer
        Array mapping indices of 'needle' "by" keys to the index of the same key
        from 'haystack', if the key also exists there.
    needle_idx_in_haystack : np.ndarray of integer
        Output array, which, after this function returns, will contain an integer fancy index
        where each element indicates an index within `haystack` where
        The shape must be the same as `needles_igroup` and the dtype should match `haystack_igroup`.
    haystack_first_unsorted_idx_in_group, needle_first_unsorted_idx_in_group : np.ndarray of integer
        Integer arrays (output) same length as `haystack_ncountgroup` and `needles_ncountgroup`.
        to hold indices within `haystack_2nd_key` and `needles_2nd_key`, respectively, where the
        first unsorted/out-of-order element was detected within that group (if any).
        Pre-initialize these arrays with invalid values so it's easy to detect when elements have been set.
        The arrays are checked after this function returns and are used to provide better error messages
        to users if unsorted values are detected.
    equality_tolerance : scalar, optional
    allow_exact_matches : bool, optional
        Indicates whether exact matches are allowed for the 'on' values.
        Specify ``None`` for ``False`` and ``True`` for ``True``;
        this leverages numba's type-based dispatch mechanism to specialize
        the compiled code with the appropriate comparison function.
    """
    # Get the 'invalid' value for the output array, to use when
    # we can't find a match.
    invalid_haystack_idx = get_invalid(needle_idx_in_haystack[0])

    for needles_key_idx in nb.prange(len(needle_key_to_haystack_key)):  # pylint: disable=not-an-iterable
        # Add one (1) to convert from the key index to the group index,
        # because group 0 represents the null/NA group.
        needle_group_offset = needles_ifirstgroup[needles_key_idx + 1]
        needle_group_length = needles_ncountgroup[needles_key_idx + 1]

        # Does this 'needles' KEY (1st key only) match a 'haystack' KEY (1st key only)?
        # If so, what's the corresponding 'haystack' group index?
        haystack_key_idx = needle_key_to_haystack_key[needles_key_idx]
        if haystack_key_idx < 0:
            # This needle doesn't exist in 'haystack'.
            # Set the values of all elements in 'needle_idx_in_haystack' that match this group to the invalid value for the array.
            needle_idx_in_haystack[
                needles_igroup[needle_group_offset : needle_group_offset + needle_group_length]
            ] = invalid_haystack_idx

            # Continue processing the next group.
            continue

        # Found a match on the first key between the 'needles' and 'haystack'.
        haystack_group_offset = haystack_ifirstgroup[haystack_key_idx + 1]
        haystack_group_length = haystack_ncountgroup[haystack_key_idx + 1]

        # Initialize needle_key2_curr, haystack_key2_curr
        # to maximum VALID value for their dtype (making sure not to set to riptable invalid).
        # If we try to be clever about declaring and initializing them in the loop below, numba isn't
        # able to infer the types and can't compile the function.
        # Initializing the values here -- to the same value for 'prev' and 'curr' -- also allows
        # some checks (e.g. 'if i_needle > 0') to be elided in the loop body.
        haystack_key2_curr = get_max_valid(haystack_2nd_key)
        needle_key2_curr = get_max_valid(needles_2nd_key)

        # Walk through the needle and haystack 2nd key ("on") values in a linear,
        # mergesort-style loop. This relies on both arrays (from their respective groups)
        # being sorted; if only the 'haystack' 2nd key values were sorted (within the group)
        # we'd have to implement this as a 'searchsorted'-style binary search
        # for each needle. (Which is faster for one element, but the linear style is much
        # faster for matching a whole array of elements.)
        i_needle = needle_group_length - 1
        i_haystack = haystack_group_length - 1
        while i_needle >= 0 and i_haystack >= 0:
            # For convenience / conciseness in the rest of the loop.
            needle_curr_idx = needles_igroup[needle_group_offset + i_needle]
            haystack_curr_idx = haystack_igroup[haystack_group_offset + i_haystack]

            # Get the current 'needle' and 'haystack' elements from their 2nd keys (the 'on' arrays).
            # Check them against the previous 'needle' and 'haystack' values, respectively;
            # this in-line sortedness check on the data is more efficient (faster) than doing the
            # check separately, e.g. as a separate method.
            needle_key2_prev = needle_key2_curr
            needle_key2_curr = needles_2nd_key[needle_curr_idx]
            haystack_key2_prev = haystack_key2_curr
            haystack_key2_curr = haystack_2nd_key[haystack_curr_idx]

            # TODO: Should we also (maybe optionally?) recognize riptable integer invalid values
            #       here and consider them like NaN's w.r.t. sortedness?
            # TODO: Should we provide an option to users to ignore NaN (or any other unsorted value)?
            if not (needle_key2_prev >= needle_key2_curr):
                # Detected unsorted (and/or NaN) 'on' value in needle.
                # Save the current index of the needle (within the full array, not the group)
                # into the "unsorted index" array at this group's index,
                # then skip processing for the rest of this group because the results
                # wouldn't be meaningful anyway.
                needle_first_unsorted_idx_in_group[needles_key_idx + 1] = needle_curr_idx
                i_needle = -1
                continue

            if not (haystack_key2_prev >= haystack_key2_curr):
                # Detected unsorted (and/or NaN) 'on' value in haystack.
                haystack_first_unsorted_idx_in_group[haystack_key_idx + 1] = haystack_curr_idx
                i_haystack = -1
                continue

            # Compare the current values of the 2nd key for needle and haystack.
            if _greater_than_comparison(allow_exact_matches, haystack_key2_curr, needle_key2_curr):
                # If we haven't reached the end of the haystack, check to see if the
                # next haystack value is *also* less than (or equal to) the current needle.
                # If so, the *current* haystack value is the one we want the needle to match against,
                # because it's the smallest haystack value that's still greater than or equal to the needle.
                # Otherwise, if the next haystack is still greater than the current needle,
                # advance the haystack pointer and recurse.
                if i_haystack > 0 and _greater_than_comparison(
                    allow_exact_matches,
                    haystack_2nd_key[haystack_igroup[haystack_group_offset + i_haystack - 1]],
                    needle_key2_curr,
                ):
                    i_haystack -= 1

                else:
                    # If the current value of `needles` 2nd key is greater than or equal to the current value of `haystack` 2nd key,
                    # and their difference is *at most* the tolerance value (if specified), we found a match,
                    # so assign the current index within the haystack array to the current index of the needles array.
                    # Otherwise, the haystack value was outside the tolerance value so this isn't a match.
                    needle_idx_in_haystack[needle_curr_idx] = (
                        haystack_curr_idx
                        # TODO: Do we need to use _less_than_comparison for the equality_tolerance check here?
                        #       Or does it always allow exact matches (i.e. the difference between the values is exactly
                        #       equal to the tolerance)?
                        if equality_tolerance is None or equality_tolerance >= (haystack_key2_curr - needle_key2_curr)
                        else invalid_haystack_idx
                    )

                    i_needle -= 1

            else:
                # N.B. This assumes haystack_key2_curr < needle_key2_curr, but it might not
                # if the 2nd keys are floats and the current value is NaN.

                # This method implements the 'forward'-looking merge_asof, so if we're still at
                # the end/back (greatest index) of the haystack, we know there's no match for this needle.
                if i_haystack + 1 == haystack_group_length:
                    needle_idx_in_haystack[needle_curr_idx] = invalid_haystack_idx
                    i_needle -= 1
                else:
                    i_haystack -= 1

        # If we reached the end of the haystack and have some remaining needles,
        # they should all be greater than (or equal to) the last/greatest haystack value.
        # Set their corresponding values in the output array to point to the last haystack value.
        if i_needle >= 0:
            needle_idx_in_haystack[needles_igroup[needle_group_offset : needle_group_offset + i_needle]] = (
                0 if haystack_group_length > 0 else invalid_haystack_idx
            )


@nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache)
def _merge_asof_combine_into_nearest(
    haystack_2nd_key: np.ndarray,
    needles_2nd_key: np.ndarray,
    needle_idx_in_haystack_backward: npt.NDArray[np.integer],
    needle_idx_in_haystack_forward: npt.NDArray[np.integer],
) -> None:
    """
    Produce a 'nearest'-mode fancy index by combining fancy indices from 'backward' and 'forward'-mode merge_asof.

    This function can be used for both the "non-grouped" (no 'by' column(s) specified) and "grouped"
    merge_asof, because the 'backward' and 'forward'-mode fancy indices already incorporate the
    grouping information.

    Parameters
    ----------
    haystack_2nd_key : np.ndarray
    needles_2nd_key : np.ndarray
    needle_idx_in_haystack_backward : np.ndarray of integer
    needle_idx_in_haystack_forward : np.ndarray of integer
        Fancy index into `haystack_2nd_key` created by one of the forward-looking
        'merge_asof' implementations. This array is updated in-place and serves
        as both an input and the output of this function.

    Notes
    -----
    Requires ``len(needles_2nd_key) == len(needle_idx_in_haystack_backward) == len(needle_idx_in_haystack_forward)``.
    """
    for i in nb.prange(len(needles_2nd_key)):  # pylint: disable=not-an-iterable
        # The indices within 'haystack_2nd_key' for the backward- and forward-looking
        # matches of the current needle value.
        haystack_idx_backward = needle_idx_in_haystack_backward[i]
        haystack_idx_forward = needle_idx_in_haystack_forward[i]

        # For the current index / needle value, was there a backwards-looking match in 'haystack'?
        if is_valid(haystack_idx_backward):
            # If the index for the forward-looking match is invalid, use the backward-looking match index.
            if not is_valid(haystack_idx_forward):
                needle_idx_in_haystack_forward[i] = haystack_idx_backward
            else:
                # Matched both forwards and backwards, so determine which one is closer.
                needle_key2_curr = needles_2nd_key[i]

                # N.B. While computing the differences here, we use the contract between the
                # needle and haystack values in backwards (haystack <= needle) and
                # forwards (haystack >= needle) so both diffs will be >= 0.
                diff_backward = needle_key2_curr - haystack_2nd_key[haystack_idx_backward]
                diff_forward = haystack_2nd_key[haystack_idx_forward] - needle_key2_curr

                # If the 'haystack' value from the backwards-looking match is closer
                # to the 'needle' value than the 'haystack' value from the forwards-looking
                # match, update the fancy-index array to use the backwards-looking index
                # for this row.
                # NOTE: For backwards-compatibility, 'backwards' wins any ties,
                # so this comparison uses __le__ instead of __lt__.
                if diff_backward <= diff_forward:
                    needle_idx_in_haystack_forward[i] = haystack_idx_backward


def _fancyindex_dtype_from_shape(shape: Union[int, Tuple[int, ...]]) -> np.dtype:
    # Get max length of any dimension.
    max_dim_len = max(shape) if isinstance(shape, (tuple, list, np.ndarray)) else shape

    # Use -(... + 1) approach with np.min_scalar_type()
    return np.min_scalar_type(-(max_dim_len + 1))


def _create_merge_asof_fancy_indices(
    left_on: np.ndarray,
    right_on: np.ndarray,
    tolerance: Optional[Union[int, float, "timedelta", TimeSpanScalar]] = None,
    allow_exact_matches: bool = True,
    direction: Literal["backward", "forward", "nearest"] = "backward",
) -> JoinIndices:
    """
    Create two fancy indices -- for the left and right Datasets, respectively -- to be used to index the columns
    of those Datasets to produce the joined result Dataset.

    This function is for the non-grouped version of `merge_asof`; that is, where no 'by' column(s)
    have been specified.

    Parameters
    ----------
    left_on
    right_on
    tolerance : integer or float or Timedelta, optional, default None
        Tolerance allowed when performing the 'asof' part of the merge; whenever a row from
        `left` doesn't have a key in `right` within this distance or less, that row will have
        a null/missing/NA value for any columns from the `right` Dataset which appear in the
        merged result.
    allow_exact_matches : boolean, default True
        - If True, allow matching with the same 'on' value
          (i.e. less-than-or-equal-to / greater-than-or-equal-to)
        - If False, don't match the same 'on' value
          (i.e., strictly less-than / strictly greater-than)
    direction : {'backward', 'forward', or 'nearest'}, default 'backward'
        Whether to search for prior, subsequent, or closest matches.

    Returns
    -------
    JoinIndices
        A `JoinIndices` instance containing the results from the join algorithm.

    Notes
    -----
    No 'by' here -- this is a pure mergesort-type join along the 'on' columns.
    Basically, it's a specialized 'searchsorted' where BOTH arrays (needles and haystack)
    are sorted (not just haystack, as with normal searchsorted). This means the implementation
    can operate in O(n+m) time instead of O(n*log(m)).

    TODO: Consider allowing the ``tolerance`` parameter to also be a 1D array/sequence whose length is the same
          as the number of groups (or keys?) in the ``left`` dataset. That allows a per-group tolerance to be specified
          if needed.
    """
    # Allocate the output array.
    # TODO: Should be able to use rt.empty here and have it completely overwritten by the numba-based
    #       implementation, but for some reason it's not. So for now workaround by initializing the
    #       array with all elements set to the default value.
    right_fancyindex_dtype = _fancyindex_dtype_from_shape(right_on.shape)
    right_fancyindex = empty(left_on.shape, dtype=right_fancyindex_dtype)
    right_fancyindex[:] = right_fancyindex.inv

    # Allocate arrays the numba function can use to notify us that the
    # 'on' data isn't completely sorted within the arrays.
    haystack_first_unsorted_idx = empty(1, dtype=np.int64)
    haystack_first_unsorted_idx[0] = haystack_first_unsorted_idx.inv
    needle_first_unsorted_idx = empty(1, dtype=np.int64)
    needle_first_unsorted_idx[0] = needle_first_unsorted_idx.inv

    # Dispatch to numba-based implementation based on 'direction'.
    if direction == "backward":
        _merge_asof_backward(
            right_on,
            left_on,
            right_fancyindex,
            haystack_first_unsorted_idx,
            needle_first_unsorted_idx,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

    elif direction == "forward":
        _merge_asof_forward(
            right_on,
            left_on,
            right_fancyindex,
            haystack_first_unsorted_idx,
            needle_first_unsorted_idx,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

    elif direction == "nearest":
        # TEMP: For now, implement the 'nearest' mode by running both 'backwards' and 'forwards' modes
        #       then combining the results. We can implement a specialized function for it later to
        #       improve the performance.

        # Run both a backwards-looking and forwards-looking 'merge_asof'.
        right_fancyindex_backward = right_fancyindex
        _merge_asof_backward(
            right_on,
            left_on,
            right_fancyindex_backward,
            haystack_first_unsorted_idx,
            needle_first_unsorted_idx,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

        right_fancyindex_forward = empty(right_fancyindex_backward.shape, dtype=right_fancyindex_backward.dtype)
        right_fancyindex_forward[:] = right_fancyindex_forward.inv

        _merge_asof_forward(
            right_on,
            left_on,
            right_fancyindex_forward,
            haystack_first_unsorted_idx,
            needle_first_unsorted_idx,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

        # Combine the backwards-looking and forwards-looking results into a single fancy-index
        # based on which 'haystack' value is nearer for each row in 'needles'.
        # Note 'backward' wins ties.
        _merge_asof_combine_into_nearest(right_on, left_on, right_fancyindex_backward, right_fancyindex_forward)
        right_fancyindex = right_fancyindex_forward

    else:
        raise ValueError(f"Invalid direction value '{direction}' specified.")

    # Check whether any groups weren't sorted; if we find one or more (in either side),
    # raise an exception with the keys of some unsorted groups.
    if np.any(isnotnan(haystack_first_unsorted_idx)):
        raise ValueError("The 'on' column for the 'right' Dataset contains one or more unsorted elements.")
    elif np.any(isnotnan(needle_first_unsorted_idx)):
        raise ValueError("The 'on' column for the 'left' Dataset contains one or more unsorted elements.")

    # Create and return a JoinIndices with the right-fancyindex.
    return JoinIndices(None, right_fancyindex)


def _create_merge_asof_grouped_fancy_indices(
    left_by_keygroup: "Grouping",
    right_by_keygroup: "Grouping",
    left_on: np.ndarray,
    right_on: np.ndarray,
    tolerance: Optional[Union[int, float, "timedelta", TimeSpanScalar]] = None,
    allow_exact_matches: bool = True,
    direction: Literal["backward", "forward", "nearest"] = "backward",
) -> JoinIndices:
    """
    Create two fancy indices -- for the left and right Datasets, respectively -- to be used to index the columns
    of those Datasets to produce the joined result Dataset.

    Parameters
    ----------
    left_by_keygroup : Grouping
        A `Grouping` instance representing the 'by' key(s) from the 'left' Dataset.
    right_by_keygroup : Grouping
        A `Grouping` instance representing the 'by' key(s) from the 'right' Dataset.
    left_on
    right_on
    tolerance : integer or float or Timedelta, optional, default None
        Tolerance allowed when performing the 'asof' part of the merge; whenever a row from
        `left` doesn't have a key in `right` within this distance or less, that row will have
        a null/missing/NA value for any columns from the `right` Dataset which appear in the
        merged result.
    allow_exact_matches : boolean, default True
        - If True, allow matching with the same 'on' value
          (i.e. less-than-or-equal-to / greater-than-or-equal-to)
        - If False, don't match the same 'on' value
          (i.e., strictly less-than / strictly greater-than)
    direction : {'backward', 'forward', or 'nearest'}, default 'backward'
        Whether to search for prior, subsequent, or closest matches.

    Returns
    -------
    JoinIndices
        A `JoinIndices` instance containing the results from the join algorithm.

    Notes
    -----
    TODO: Consider allowing the ``tolerance`` parameter to also be a 1D array/sequence whose length is the same
          as the number of groups (or keys?) in the ``left`` dataset. That allows a per-group tolerance to be specified
          if needed.
    """
    # Get the unique values from the Groupings.
    # TODO: If any of the keys in the tuples below are Categorical, need to expand them back;
    #       current rt.ismember() won't handle it correctly, at least if they are one of
    #       multiple keys in the Grouping.
    left_grouping_gbkey = _gbkeys_extract(left_by_keygroup)
    right_grouping_gbkey = _gbkeys_extract(right_by_keygroup)

    # Determine which keys of the 'left_by' Grouping are also present in the 'right_by'
    # Grouping, and create a mapping from 'left_by' key indices to 'right_by' key indices.
    # N.B. we do this operation on the gbkeys of each grouping; the number of uniques in each
    # key will be smaller-than-or-equal to the length of the original data column, and in most
    # cases will be significantly smaller (so using ismember on the gbkeys will be significantly faster).
    # TODO: Pass 'hint_size' here to speed up the 'ismember' call?
    _, left_by_keyid_to_right_by_keyid_map = ismember(left_grouping_gbkey, right_grouping_gbkey)

    # Allocate the output array.
    # TODO: Should be able to use rt.empty here and have it completely overwritten by the numba-based
    #       implementation, but for some reason it's not. So for now workaround by initializing the
    #       array with all elements set to the default value.
    right_fancyindex = empty(left_by_keygroup.ikey.shape, dtype=right_by_keygroup.igroup.dtype)
    right_fancyindex[:] = right_fancyindex.inv

    # Allocate arrays the numba function can use to notify us that the
    # 'on' data isn't completely sorted within some group(s).
    haystack_first_unsorted_idx_in_group = full_like(right_by_keygroup.ifirstgroup, right_by_keygroup.ifirstgroup.inv)
    needle_first_unsorted_idx_in_group = full_like(left_by_keygroup.ifirstgroup, left_by_keygroup.ifirstgroup.inv)

    # Dispatch to numba-based implementation based on 'direction'.
    if direction == "backward":
        _merge_asof_grouped_backward(
            right_by_keygroup.igroup,
            right_by_keygroup.ifirstgroup,
            right_by_keygroup.ncountgroup,
            right_on,
            left_by_keygroup.igroup,
            left_by_keygroup.ifirstgroup,
            left_by_keygroup.ncountgroup,
            left_on,
            left_by_keyid_to_right_by_keyid_map,
            right_fancyindex,
            haystack_first_unsorted_idx_in_group,
            needle_first_unsorted_idx_in_group,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

    elif direction == "forward":
        _merge_asof_grouped_forward(
            right_by_keygroup.igroup,
            right_by_keygroup.ifirstgroup,
            right_by_keygroup.ncountgroup,
            right_on,
            left_by_keygroup.igroup,
            left_by_keygroup.ifirstgroup,
            left_by_keygroup.ncountgroup,
            left_on,
            left_by_keyid_to_right_by_keyid_map,
            right_fancyindex,
            haystack_first_unsorted_idx_in_group,
            needle_first_unsorted_idx_in_group,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

    elif direction == "nearest":
        # TEMP: For now, implement the 'nearest' mode by running both 'backwards' and 'forwards' modes
        #       then combining the results. We can implement a specialized function for it later to
        #       improve the performance.

        # Run both a backwards-looking and forwards-looking 'merge_asof'.
        right_fancyindex_backward = right_fancyindex
        _merge_asof_grouped_backward(
            right_by_keygroup.igroup,
            right_by_keygroup.ifirstgroup,
            right_by_keygroup.ncountgroup,
            right_on,
            left_by_keygroup.igroup,
            left_by_keygroup.ifirstgroup,
            left_by_keygroup.ncountgroup,
            left_on,
            left_by_keyid_to_right_by_keyid_map,
            right_fancyindex_backward,
            haystack_first_unsorted_idx_in_group,
            needle_first_unsorted_idx_in_group,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

        right_fancyindex_forward = empty(right_fancyindex_backward.shape, dtype=right_fancyindex_backward.dtype)
        right_fancyindex_forward[:] = right_fancyindex_forward.inv

        _merge_asof_grouped_forward(
            right_by_keygroup.igroup,
            right_by_keygroup.ifirstgroup,
            right_by_keygroup.ncountgroup,
            right_on,
            left_by_keygroup.igroup,
            left_by_keygroup.ifirstgroup,
            left_by_keygroup.ncountgroup,
            left_on,
            left_by_keyid_to_right_by_keyid_map,
            right_fancyindex_forward,
            # Re-use these for the forward call. If unsortedness is detected, it'll be
            # at the same index in both forwards and backwards mode.
            haystack_first_unsorted_idx_in_group,
            needle_first_unsorted_idx_in_group,
            tolerance,
            # Specify None if not allowing exact matches,
            # so numba can specialize the function with the correct operator.
            # (numba will specialize on types, but not on values.)
            True if allow_exact_matches else None,
        )

        # Combine the backwards-looking and forwards-looking results into a single fancy-index
        # based on which 'haystack' value is nearer for each row in 'needles'.
        # Note 'backward' wins ties.
        _merge_asof_combine_into_nearest(right_on, left_on, right_fancyindex_backward, right_fancyindex_forward)
        right_fancyindex = right_fancyindex_forward

    else:
        raise ValueError(f"Invalid direction value '{direction}' specified.")

    # Check whether any groups weren't sorted; if we find one or more (in either side),
    # raise an exception with the keys of some unsorted groups.
    if np.any(isnotnan(haystack_first_unsorted_idx_in_group)):
        raise ValueError("The 'on' column isn't sorted within one or more unsorted 'by' groups of the 'right' Dataset.")
    elif np.any(isnotnan(needle_first_unsorted_idx_in_group)):
        raise ValueError("The 'on' column isn't sorted within one or more unsorted 'by' groups of the 'left' Dataset.")

    # Create and return a JoinIndices with the right-fancyindex.
    return JoinIndices(None, right_fancyindex)


def _validate_and_normalize_tolerance(
    left_on: np.ndarray, right_on: np.ndarray, tolerance: Union[int, float, "timedelta", TimeSpanScalar]
) -> Any:
    """
    Validate the tolerance value is compatible with the 'on' columns for an "as-of" merge.

    The tolerance value may be converted to another scalar type before returning, if it's
    necessary for compatibility with the 'on' columns in the generated code. For example,
    a `TimeSpanScalar` may be converted to a `DateTimeNanoScalar` to ensure the comparison
    works correctly.

    Parameters
    ----------
    left_on, right_on
    tolerance

    Returns
    -------
    validated_tolerance
    """
    # If 'tolerance' is a value of a signed type, don't allow negative values;
    # the 'as-of' logic won't work correctly if given a negative value.
    negative_tol_errmsg = "Negative tolerance values not supported for as-of merges."
    if isinstance(tolerance, float) or (
        isinstance(tolerance, np.generic) and np.issubdtype(tolerance.dtype, np.floating)
    ):
        if tolerance < 0 or isnan(tolerance):
            raise ValueError(negative_tol_errmsg)

        # Validate the 'left_on' and 'right_on' columns are np.ndarray or FastArray (but no derived FA types).
        if (type(left_on) != np.ndarray and type(left_on) != FastArray) or (
            type(right_on) != np.ndarray and type(right_on) != FastArray
        ):
            raise ValueError(f"`{type(tolerance)}`-typed tolerance values not supported for the given 'on' columns.")

        # Determine which dtype is compatible between the two 'on' columns,
        # then convert the tolerance value to a scalar of that type.
        # This prevents the tolerance value from changing the types that'd be inferred
        # by numba for the generated code; if the tolerance value is large enough to
        # require a larger type than the 'on' columns, it might make sense not to allow it?
        unified_on_type: np.dtype = np.promote_types(left_on.dtype, right_on.dtype)
        return np.array(tolerance, dtype=unified_on_type)[()]

    elif isinstance(tolerance, int) or (
        isinstance(tolerance, np.generic) and np.issubdtype(tolerance.dtype, np.integer)
    ):
        if tolerance < 0 or isnan(tolerance):
            raise ValueError(negative_tol_errmsg)

        # Validate the 'left_on' and 'right_on' columns are np.ndarray or FastArray (but no derived FA types)
        if (type(left_on) != np.ndarray and type(left_on) != FastArray) or (
            type(right_on) != np.ndarray and type(right_on) != FastArray
        ):
            raise ValueError(f"`{type(tolerance)}`-typed tolerance values not supported for the given 'on' columns.")

        # Determine which dtype is compatible between the two 'on' columns,
        # then convert the tolerance value to a scalar of that type.
        # This prevents the tolerance value from changing the types that'd be inferred
        # by numba for the generated code; if the tolerance value is large enough to
        # require a larger type than the 'on' columns, it might make sense not to allow it?
        unified_on_type: np.dtype = np.promote_types(left_on.dtype, right_on.dtype)
        return np.array(tolerance, dtype=unified_on_type)[()]

    elif isinstance(tolerance, TimeSpanScalar):
        if tolerance < TimeSpanScalar(0):
            raise ValueError(negative_tol_errmsg)

        if isinstance(left_on, DateTimeNano):
            # Convert to a DateTimeNanoScalar; this is a bit weird (since we'd usually want
            # the difference of two DateTimeNano values to be a TimeSpanScalar), but due to
            # the internal representation of these types we need a DateTimeNanoScalar for the
            # JIT-compiled code to work correctly.
            return DateTimeNanoScalar(tolerance.nanoseconds)

        else:
            raise ValueError(
                f"Can't determine how to use a `{type(tolerance).__qualname__}`-type tolerance value with an 'on' column of type '{type(left_on).__name__}' and dtype '{left_on.dtype}'."
            )

    elif isinstance(tolerance, timedelta):
        if tolerance < timedelta(0):
            raise ValueError(negative_tol_errmsg)

        if isinstance(left_on, DateTimeNano):
            # The range of datetime.timedelta is larger than can be represented
            # by a 64-bit integer (signed or not) of epoch-nanoseconds.
            # If outside the range, raise an exception since we can't convert.
            if tolerance.days > 106751:  # roughly 106751.9 days in 2**63 nanoseconds
                raise ValueError("Tolerance is too large to be used with `DateTimeNano`-typed 'on' columns.")

            # Convert to nanoseconds.
            # N.B. The conversion used here is only safe for non-negative timedelta values.
            return DateTimeNanoScalar(
                (((tolerance.days * 86_400) + tolerance.seconds) * 1_000_000_000) + (tolerance.microseconds * 1_000)
            )

        else:
            raise ValueError(
                f"Can't determine how to use a `{type(tolerance).__qualname__}`-type tolerance value with an 'on' column of type '{type(left_on).__name__}' and dtype '{left_on.dtype}'."
            )

    else:
        raise ValueError(f"Tolerance value has unsupported type '{type(tolerance).__qualname__}'.")


def merge_asof2(
    left: "Dataset",
    right: "Dataset",
    on: Optional[Union[str, Tuple[str, str]]] = None,
    left_on: Optional[str] = None,
    right_on: Optional[str] = None,
    by: Optional[Union[str, Tuple[str, str], List[Union[str, Tuple[str, str]]]]] = None,
    left_by: Optional[Union[str, List[str]]] = None,
    right_by: Optional[Union[str, List[str]]] = None,
    suffixes: Optional[Tuple[str, str]] = None,
    copy: bool = True,
    columns_left: Optional[Union[str, List[str]]] = None,
    columns_right: Optional[Union[str, List[str]]] = None,
    *,
    tolerance: Optional[Union[int, float, "timedelta", TimeSpanScalar]] = None,
    allow_exact_matches: bool = True,
    direction: Literal["backward", "forward", "nearest"] = "backward",
    matched_on: Union[bool, str] = False,
    **kwargs,
) -> "Dataset":
    """
    Perform an as-of merge. This is similar to a left-join except that we match on nearest key rather than equal keys.

    Both Datasets must be sorted (ascending) by the 'on' column. When 'by' columns are specified,
    the 'on' column for each Dataset only needs to be sorted (ascending) within each unique 'key'
    of the 'by' columns. Sorting the entire `Dataset` by the 'on' column also meets this requirement,
    but some `Datasets` may have an 'on' column which is already pre-sorted within each 'by' key,
    in which case no additional sorting is required.

    For each row in the left Dataset:
      - A "backward" search selects the last row in the right Dataset whose
        'on' key is less than or equal to the left's key.
      - A "forward" search selects the first row in the right Dataset whose
        'on' key is greater than or equal to the left's key.
      - A "nearest" search selects the row in the right Dataset whose 'on'
        key is closest in absolute distance to the left's key.

    Optionally match on equivalent keys with 'by' before searching with 'on'.

    Parameters
    ----------
    left : Dataset
        Left Dataset
    right : Dataset
        Right Dataset
    on : str
        Column name to join on. Must be found in both the `left` and `right` Datasets.
        This column in both left and right Datasets MUST be ordered.
        Furthermore this must be a numeric column, such as datetimelike,
        integer, or float. Either `on` or `left_on`/`right_on` must be specified.
    left_on : str or list of str, optional
        Column name to join on in `left` Dataset.
    right_on : str or list of str, optional
        Column name to join on in `right` Dataset.
    by : str or (str, str) or list of str or list of (str, str), optional
        Column name or list of column names. Match on these columns before
        performing merge operation.
    left_by : str or list of str, optional
        Column names to match on in the left Dataset.
    right_by : str or list of str, optional
        Column names to match on in the right Dataset.
    suffixes : (str, str), optional, default None
        Suffix to apply to overlapping column names in the left and right
        side, respectively.
    copy: bool, default True
        If False, avoid copying data when possible; this can reduce memory usage
        but users must be aware that data can be shared between `left` and/or `right`
        and the Dataset returned by this function.
    columns_left : str or list of str, optional
        Column names to include in the merge from `left`, defaults to None which causes all columns to be included.
    columns_right : str or list of str, optional
        Column names to include in the merge from `right`, defaults to None which causes all columns to be included.
    tolerance : integer or float or Timedelta, optional, default None
        Tolerance allowed when performing the 'asof' part of the merge; whenever a row from
        `left` doesn't have a key in `right` within this distance or less, that row will have
        a null/missing/NA value for any columns from the `right` Dataset which appear in the
        merged result.
    allow_exact_matches : boolean, default True
        - If True, allow matching with the same 'on' value
          (i.e. less-than-or-equal-to / greater-than-or-equal-to)
        - If False, don't match the same 'on' value
          (i.e., strictly less-than / strictly greater-than)
    direction : {'backward', 'forward', or 'nearest'}, default 'backward'
        Whether to search for prior, subsequent, or closest matches.
    matched_on : bool or str, default False
        If set to True or a string, an additional column is added to the result;
        for each row, it contains the value from the `on` column in `right` that was matched.
        When True, the column will use the default name 'matched_on'; specify a string
        to explicitly name the column.

    Returns
    -------
    merged : Dataset

    Raises
    ------
    ValueError
        The `on`, `left_on`, or `right_on` columns are not sorted in ascending order
        within one or more keys of the `by` columns.

    See Also
    --------
    riptable.merge_asof

    Notes
    -----
    TODO: Consider allowing the ``tolerance`` parameter to also be a 1D array/sequence whose length is the same
          as the number of groups (or keys?) in the ``left`` dataset. That allows a per-group tolerance to be specified
          if needed.

    Examples
    --------
    >>> left = rt.Dataset({'a': [1, 5, 10], 'left_val': ['a', 'b', 'c']})
    >>> left
    #    a   left_val
    -   --   --------
    0    1   a
    1    5   b
    2   10   c
    >>> right = rt.Dataset({'a': [1, 2, 3, 6, 7],
    ...                       'right_val': [1, 2, 3, 6, 7]})
    >>> right
    #   a   right_val
    -   -   ---------
    0   1           1
    1   2           2
    2   3           3
    3   6           6
    4   7           7
    >>> rt.merge_asof(left, right, on='a')
    #   a_x   left_val   a_y   right_val
    -   ---   --------   ---   ---------
    0     1   a            1           1
    1     5   b            3           3
    2    10   c            7           7

    >>> rt.merge_asof(left, right, on='a', allow_exact_matches=False)
     #   a_x   left_val   a_y   right_val
    -   ---   --------   ---   ---------
    0     1   a          Inv         Inv
    1     5   b            3           3
    2    10   c            7           7

    >>> rt.merge_asof(left, right, on='a', direction='forward')
    #   a_x   left_val   a_y   right_val
    -   ---   --------   ---   ---------
    0     1   a            1           1
    1     5   b            6           6
    2    10   c          Inv         Inv

    Here is a real-world time-series example

    >>> quotes
    #                   time  ticker   Bid   Ask
    -   --------------------  ------  ----  ----
    0   20191015 09:45:57.09    AAPL  3.40  3.50
    1   20191015 11:35:09.76    AAPL  3.45  3.55
    2   20191015 12:02:27.11    AAPL  3.50  3.60
    3   20191015 12:43:13.73    MSFT  2.85  2.95
    4   20191015 14:32:11.18    MSFT  2.90  3.00

    >>> trades
    #                   time  ticker  TradePrice  TradeSize
    -   --------------------  ------  ----------  ---------
    0   20191015 10:03:24.73    AAPL        3.45       1.00
    1   20191015 10:41:22.79    MSFT        2.85       1.00
    2   20191015 10:41:35.69    MSFT        2.86       1.00
    3   20191015 11:04:32.55    AAPL        3.47       1.00
    4   20191015 11:44:35.63    MSFT        2.91       1.00
    5   20191015 12:26:17.68    AAPL        3.55       1.00
    6   20191015 14:24:10.93    MSFT        2.98       1.00
    7   20191015 15:45:13.41    AAPL        3.60       7.00
    8   20191015 15:50:42.53    AAPL        3.58       1.00
    9   20191015 15:53:59.60    AAPL        3.56       5.00

    >>> rt.merge_asof(trades, quotes, on='time', by='ticker')
    #                 time_x   ticker_x  TradePrice   TradeSize                 time_y   ticker_y    Bid    Ask
    -   --------------------  ---------  ----------   ---------   --------------------   --------   ----   ----
    0   20191015 10:03:24.73       AAPL        3.45        1.00   20191015 09:45:57.09       AAPL   3.40   3.50
    1   20191015 10:41:22.79       MSFT        2.85        1.00                    Inv   Filtered    nan    nan
    2   20191015 10:41:35.69       MSFT        2.86        1.00                    Inv   Filtered    nan    nan
    3   20191015 11:04:32.55       AAPL        3.47        1.00   20191015 09:45:57.09       AAPL   3.40   3.50
    4   20191015 11:44:35.63       MSFT        2.91        1.00                    Inv   Filtered    nan    nan
    5   20191015 12:26:17.68       AAPL        3.55        1.00   20191015 12:02:27.11       AAPL   3.50   3.60
    6   20191015 14:24:10.93       MSFT        2.98        1.00   20191015 12:43:13.73       MSFT   2.85   2.95
    7   20191015 15:45:13.41       AAPL        3.60        7.00   20191015 12:02:27.11       AAPL   3.50   3.60
    8   20191015 15:50:42.53       AAPL        3.58        1.00   20191015 12:02:27.11       AAPL   3.50   3.60
    9   20191015 15:53:59.60       AAPL        3.56        5.00   20191015 12:02:27.11       AAPL   3.50   3.60

    only merge with the forward quotes

    >>> rt.merge_asof(trades, quotes, on='time', by='ticker', direction='forward')
    #                 time_x  ticker_x  TradePrice  TradeSize                time_y  ticker_y   Bid   Ask
    -   --------------------  --------  ----------  ---------  --------------------  --------  ----  ----
    0   20191015 10:03:24.73      AAPL        3.45       1.00  20191015 11:35:09.76      AAPL  3.45  3.55
    1   20191015 10:41:22.79      MSFT        2.85       1.00  20191015 12:43:13.73      MSFT  2.85  2.95
    2   20191015 10:41:35.69      MSFT        2.86       1.00  20191015 12:43:13.73      MSFT  2.85  2.95
    3   20191015 11:04:32.55      AAPL        3.47       1.00  20191015 11:35:09.76      AAPL  3.45  3.55
    4   20191015 11:44:35.63      MSFT        2.91       1.00  20191015 12:43:13.73      MSFT  2.85  2.95
    6   20191015 14:24:10.93      MSFT        2.98       1.00  20191015 14:32:11.18      MSFT  2.90  3.00
    7   20191015 15:45:13.41      AAPL        3.60       7.00                   Inv  Filtered   nan   nan
    8   20191015 15:50:42.53      AAPL        3.58       1.00                   Inv  Filtered   nan   nan
    9   20191015 15:53:59.60      AAPL        3.56       5.00                   Inv  Filtered   nan   nan
    5   20191015 12:26:17.68      AAPL        3.55       1.00                   Inv  Filtered   nan   nan
    """
    # Process keyword arguments.
    left_index = bool(kwargs.pop("left_index", False))
    right_index = bool(kwargs.pop("right_index", False))
    _ = bool(kwargs.pop("check_sorted", True))
    _ = bool(kwargs.pop("verbose", False))
    if kwargs:
        # There were remaining keyword args passed here which we don't understand.
        first_kwarg = next(iter(kwargs.keys()))
        raise ValueError(f"This function does not support the kwarg '{first_kwarg}'.")

    if left_index or right_index:
        # Emit warning about 'left_index' and 'right_index' only being present for compatibility
        # with the pandas merge signature. They don't actually do anything in riptable since our
        # indexing is external (not internal) to Datasets.
        warnings.warn(
            "The 'left_index' and 'right_index' parameters are only present for pandas compatibility. They are not applicable for riptable and will have no effect."
        )

    # Validate the 'direction' argument.
    if direction not in ("forward", "backward", "nearest"):
        raise ValueError(f"`{direction}` is not a valid value for the 'direction' parameter.")

    # Validate and normalize the 'on' column name for each Dataset.
    if left_on is None:
        if on is None:
            raise ValueError("The `on` and `left_on` parameters cannot both be None.")
        elif isinstance(on, str):
            left_on = on
        elif isinstance(on, bytes):
            left_on = on.decode()
        else:
            # Assume `on` is a 2-tuple of strings.
            left_on = on[0]
    else:
        # 'on' and 'left_on' are not allowed to be specified together, as it's unclear which one should take precedence.
        # If we want to define which one does take precedence, we could drop this down to
        # a Warning to let the user know part of what they've specified will be ignored.
        if on is not None:
            raise ValueError(
                "The `on` and `left_on` parameters cannot be specified together; exactly one of them should be specified."
            )

    if right_on is None:
        if on is None:
            raise ValueError("The `on` and `right_on` parameters cannot both be None.")
        elif isinstance(on, str):
            right_on = on
        elif isinstance(on, bytes):
            right_on = on.decode()
        else:
            # Assume `on` is a 2-tuple of strings.
            right_on = on[1]
    else:
        # 'on' and 'right_on' are not allowed to be specified together, as it's unclear which one should take precedence.
        # If we want to define which one does take precedence, we could drop this down to
        # a Warning to let the user know part of what they've specified will be ignored.
        if on is not None:
            raise ValueError(
                "The `on` and `right_on` parameters cannot be specified together; exactly one of them should be specified."
            )

    # Validate and normalize the 'by' column name lists for each Dataset.
    # Note that for 'merge_asof', specifying any 'by' columns are optional -- unlike the 'on' columns, which are required.
    left_by = _extract_on_columns(by, left_by, True, "by", is_optional=True)
    right_by = _extract_on_columns(by, right_by, False, "by", is_optional=True)

    #
    # TODO: Disallow a column to be specified as both an 'on' and 'by' column (for each Dataset)?
    #       Using the same column for both means the 'on' column would effectively be ignored since
    #       the 'by' part is already doing an exact match (equi-join). If that's what someone wants,
    #       seems like they should be using merge_lookup() or merge2() instead.
    #

    # Normalize 'columns_left' and 'columns_right' first to simplify some logic later on
    # (by allowing us to assume they're a non-optional-but-maybe-empty List[str]).
    columns_left = _normalize_selected_columns(left, columns_left)
    columns_right = _normalize_selected_columns(right, columns_right)

    # PERF: Revisit this -- it could be made faster if ItemContainer.keys() returned a set-like object such as KeysView instead of a list; then we wouldn't need to create the sets here.
    left_keyset = set(left.keys())
    _require_columns_present(left_keyset, "left", "left_on", [left_on])
    _require_columns_present(left_keyset, "left", "left_by", left_by)
    _require_columns_present(
        left_keyset, "left", "columns_left", columns_left
    )  # PERF: Fix this -- if columns_left isn't populated initially it'll be normalized to the whole keyset above so this call is irrelevant
    right_keyset = set(right.keys())
    _require_columns_present(right_keyset, "right", "right_on", [right_on])
    _require_columns_present(right_keyset, "right", "right_by", right_by)
    _require_columns_present(
        right_keyset, "right", "columns_right", columns_right
    )  # PERF: Fix this -- if columns_right isn't populated initially it'll be normalized to the whole keyset above so this call is irrelevant

    # Make sure there aren't any column name collision _before_ we do the heavy lifting of merging;
    # if there are name collisions, attempt to resolve them by suffixing the colliding column names.
    # For the purposes of this validation, we combine the 'on' column and any 'by' column(s) since they're both
    # keys (it's just that the 'on' column is treated specially when the join is performed).
    left_keycol_names = [left_on]
    left_keycol_names.extend(left_by)
    right_keycol_names = [right_on]
    right_keycol_names.extend(right_by)
    col_left_tuple, col_right_tuple, intersection_cols = _construct_colname_mapping(
        left_keycol_names, right_keycol_names, suffixes=suffixes, columns_left=columns_left, columns_right=columns_right
    )

    # Validate the pair(s) of columns from the left and right join keys have compatible types.
    key_compat_errs = _verify_join_keys_compat(
        [left[col_name] for col_name in left_keycol_names], [right[col_name] for col_name in right_keycol_names]
    )

    if key_compat_errs:
        # If the list of errors is non-empty, we have some join-key compatibility issues.
        # Some "errors" may just be warnings; filter those out of the list and raise
        # them to notify the user.
        actual_errors: List[Exception] = []
        for err in key_compat_errs:
            if isinstance(err, Warning):
                warnings.warn(err)
            else:
                actual_errors.append(err)

        # If there are any remaining errors in the list, those are actual errors so we can't
        # proceed any further. Combine the remaining errors into a single error message then
        # raise a ValueError; in the future we might use something like a .NET AggregateException
        # to wrap multiple errors into one instead to allow for different error types.
        if actual_errors:
            flat_errs = "\n".join(
                map(str, actual_errors)
            )  # N.B. this is because it's disallowed to use backslashes inside f-string curly braces
            raise ValueError(f"Found one or more compatibility errors with the specified 'on' keys:\n{flat_errs}")

    #
    # TODO: Verify the 'on' columns are an ordered, comparable type (int, float, DateTimeNano, datetime, ordered Categorical, etc.)
    #       This is a requirement for them to be used with this function; without this, the 'asof' comparison can't be performed
    #       and will raise an exception later -- better to check here and give the user a better error message.
    #

    datasetclass = type(left)

    left_on_col = left[left_on]
    right_on_col = right[right_on]

    # If 'tolerance' is specified, verify it's a value of a type compatible with the 'on' columns.
    # The value may also be converted to a different scalar type before being passed to the numba-based impl.
    # so it'll be compatible in the JIT-compiled code.
    if tolerance is not None:
        tolerance = _validate_and_normalize_tolerance(left_on_col, right_on_col, tolerance)

    # Construct the Grouping object for each of the join keys.
    start = perf_counter_ns()

    # TODO: Should we always pass True for the 'force_invalids' argument below? The original intent was that we
    #       may want to be able to use the default (non-forced-invalids) behavior of Grouping for optional compatibility
    #       with rt.merge(), but that's no longer planned (users are just migrating to merge2/merge_lookup).
    #       Note, there *may* be something to passing False for the left and right grouping, when performing a
    #       left or right join, respectively -- before making the change to always pass True, verify all the
    #       optimized cases where we can return e.g. None for the fancy index aren't affected by the change.
    how = "left"
    high_card = False
    hint_size = None
    if left_by is None or len(left_by) == 0:
        left_grouping = None
    else:
        left_grouping, _ = _get_or_create_keygroup(
            [left[col_name] for col_name in left_by], 0, how != "left", False, high_card, hint_size
        )
    if right_by is None or len(right_by) == 0:
        right_grouping = None
    else:
        right_grouping, _ = _get_or_create_keygroup(
            [right[col_name] for col_name in right_by], 1, how != "right", how == "outer", high_card, hint_size
        )

    if _logger.isEnabledFor(logging.DEBUG):
        delta = perf_counter_ns() - start
        _logger.debug("Grouping creation complete.", extra={"elapsed_nanos": delta})

    # Construct fancy indices for the left/right Datasets; these will be used to index into
    # columns of the respective datasets to produce new arrays/columns for the merged Dataset.
    if left_grouping is None:
        join_indices = _create_merge_asof_fancy_indices(
            left_on_col,
            right_on_col,
            tolerance=tolerance,
            allow_exact_matches=allow_exact_matches,
            direction=direction,
        )
    else:
        join_indices = _create_merge_asof_grouped_fancy_indices(
            left_grouping,
            right_grouping,
            left_on_col,
            right_on_col,
            tolerance=tolerance,
            allow_exact_matches=allow_exact_matches,
            direction=direction,
        )

    left_fancyindex = join_indices.left_index  # Should always be None for this type of join
    right_fancyindex = join_indices.right_index

    # Begin creating the column data for the 'merged' Dataset.
    out: Dict[str, FastArray] = {}
    start = perf_counter_ns()

    def readonly_array_wrapper(arr: FastArray) -> FastArray:
        """Create a read-only view of an array."""
        new_arr = FastArray(arr)
        new_arr.flags.writeable = False
        return new_arr

    def array_copy(arr: FastArray) -> FastArray:
        return arr.copy()

    # Based on the 'copy' flag, determine which function to use in cases where
    # an input column will contain the same data in the output Dataset.
    array_copier = array_copy if copy else readonly_array_wrapper

    # Fetch/transform columns which overlap between the two Datasets (i.e. columns with these
    # names exist in both Datasets).
    if intersection_cols:
        # If we're missing one of the fancy indices, it means we can just copy the columns
        # over from the corresponding dataset to the result. We still allocate a new array
        # object as a read-only view of the original; this allows the name to be different
        # in the result (e.g. if changed later) than the original.
        if left_fancyindex is None:
            for field in intersection_cols:
                out[field] = array_copier(left[field])
        elif right_fancyindex is None:
            for field in intersection_cols:
                out[field] = array_copier(right[field])

        else:
            # For any columns which overlap (i.e. same name) in the two Datasets,
            # fetch them from the left Dataset; we don't support right-joins for merge_asof
            # so we don't need to handle that here as we do in merge2.
            # Applying the fancy index to transform each column to the correct shape for the resulting Dataset.
            # N.B. This is an arbitrary choice, we could just as easily default to fetching
            #      from the right Dataset instead.
            for field in intersection_cols:
                out[field] = left[field][left_fancyindex]

    if _logger.isEnabledFor(logging.DEBUG):
        delta = perf_counter_ns() - start
        _logger.debug("Transformed columns. cols='%s'", "intersection", extra={"elapsed_nanos": delta})

    # Transform the columns from the left Dataset and store to the new, merged Dataset.
    # If we don't have a left fancy index, it means the fancy index (if present) would simply create
    # a copy of the original column. Use this to optimize memory usage by just referencing (not copying)
    # the original column data in the new dataset.
    start = perf_counter_ns()
    if left_fancyindex is None:
        for old_name, new_name in zip(*col_left_tuple):
            # Wrap the original array in a read-only view; this is necessary to allow the "new" column to
            # have a different name than the original without interfering with the original array / source Dataset.
            # We make the view read-only to help protect the original data from being unexpectedly modified.
            out[new_name] = array_copier(left[old_name])
    else:
        for old_name, new_name in zip(*col_left_tuple):
            out[new_name] = left[old_name][left_fancyindex]

    if _logger.isEnabledFor(logging.DEBUG):
        delta = perf_counter_ns() - start
        _logger.debug("Transformed columns. cols='%s'", "left", extra={"elapsed_nanos": delta})

    # Transform the columns from the right Dataset and store to the new, merged Dataset.
    start = perf_counter_ns()
    if right_fancyindex is None:
        for old_name, new_name in zip(*col_right_tuple):
            # Wrap the original array in a read-only view; this is necessary to allow the "new" column to
            # have a different name than the original without interfering with the original array / source Dataset.
            # We make the view read-only to help protect the original data from being unexpectedly modified.
            out[new_name] = array_copier(right[old_name])
    else:
        for old_name, new_name in zip(*col_right_tuple):
            out[new_name] = right[old_name][right_fancyindex]

    if _logger.isEnabledFor(logging.DEBUG):
        delta = perf_counter_ns() - start
        _logger.debug("Transformed columns. cols='%s'", "right", extra={"elapsed_nanos": delta})

    # If the caller has asked for the 'matched_on' column, create it now.
    if matched_on:
        start = perf_counter_ns()

        if isinstance(matched_on, bool):
            matched_on = "matched_on"
        if matched_on in out:
            raise ValueError(f"`matched_on` column name collision with existing columns: {matched_on}")

        # Use the right_fancyindex to expand the 'on' column from the right Dataset and add it to the output.
        out[matched_on] = right_on_col[right_fancyindex]

        if _logger.isEnabledFor(logging.DEBUG):
            delta = perf_counter_ns() - start
            _logger.debug("matched_on column created.", extra={"elapsed_nanos": delta})

    return datasetclass(out)



"""Display options for formatting and displaying numeric values, datasets, and multisets."""
import os
from json import dump, load
from typing import Optional, Union

from ..rt_enum import DisplayNumberSeparator, TypeRegister
from .appdirs import user_config_dir

__all__ = [
    "DisplayOptions",
]


class DisplayOptions(object):
    # Todo alz 20191125 - fix disappearing attributes from numpy docs
    # Following the numpy class docstring guide (https://numpy.org/devdocs/docs/howto_document.html#class-docstring),
    # the following attributes and methods do not appear in the rendered HTML output.
    # Investigate sphinx.ext.napoleon interaction with class variables.
    # Property attributes are known to render correctly.
    """Provides display options for:
      1) console screen for customizing width, height, and character buffers
      2) row, columns, and styles when displaying datasets and multisets
      3) formatting headers, footers, and general string widths
      4) formatting numeric types using scientific notation or specifying precision
      5) other miscellaneous display options such as prefixing group by column names

    Attributes
    ----------
    CONSOLE_X_BUFFER : int
        Number of characters for buffer width on console display (default 30).
    CONSOLE_X_HTML : int
        Number of characters for buffer width on HTML display (default 340).
    CONSOLE_X : int
        Number of characters for console display width (default 150).
    CONSOLE_Y : int
        Number of characters for console display height (default 25).
    HTML_DISPLAY : bool
        Toggle HTML display mode (default True).
    X_PADDING : int
        Number of characters for column buffer in console display (default 4).
    Y_PADDING : int
        Number of characters for row buffer in console display (default 3).
    ROW_ALL : bool
        Toggle display of all rows for Dataset, Multiset, and Struct objects (default False).
    COL_ALL : bool
        Toggle display of all columns for Dataset, Multiset, and Struct objects (default False).
    COL_MIN : int
        Minimum columns to display for Dataset, Multiset, and Struct objects (default 1).
    COL_MAX : int
        Maximum columns to display for Dataset, Multiset, and Struct objects (default 50).
    COL_T : int
        Number of transposed rows to display, which appear as columns for Dataset, Multiset, and Struct objects (default 8).
    HEAD_ROWS : int
        Number of rows to display when calling head on a Dataset, Multiset, or Struct object (default 15).
    TAIL_ROWS : int
        Number of rows to display when calling tail on a Dataset, Multiset, or Struct object (default 15).
    MAX_ROWS : int
        Maximum number of rows to display for Dataset, Multiset, and Struct objects (default 30).
    NO_STYLES : bool
        Toggle for colors in IPython console (default False).
        Note, may be difficult to see with light background.
    COLOR_MODE : DisplayColorMode or None, optional
        Color mode for display (default None).
    CUSTOM_COMPLETION = bool
        Toggle on for attribute completion results that show in alphanumeric key, attribute, then method ordering
        for ``Dataset``, ``Multiset``, ``Struct`` (default False).

        This will override the default IPython ``Completer._complete`` to a custom variant that allows custom completer
        dispatching using the ``IPython.utils.generics.complete_object`` hook while preserving the custom ordering.

        Caution, below are the side effects when toggling this on:
        - IPython ``use_jedi`` is set to False since this approach is currently incompatible with Jedi completion
        because the code is actually evaluated on TAB.
        - IPython ``Completer._complete`` is monkey patched to change use the custom completion that is backwards compatible
        with ``Completer._complete``, but allows preserving the order.
        - As of 20191218, if ``CUSTOM_COMPLETION`` is toggled on it results in a one-time registration of custom
        attribute completion per IPython session as opposed to supporting deregistration.
    MAX_HEADER_WIDTH : int
        Maximum number of characters for header strings in a Dataset, Multiset, or Struct object (default 15).
    MAX_FOOTER_WIDTH : int
        Maximum number of characters for footer strings in Dataset, Multiset, or Struct object  (default 15).
    MAX_STRING_WIDTH : int
        Maximum number of characters for all strings (default 15).
    PRECISION :
        Number of digits to display to the right of the decimal (default 2).
    E_PRECISION :
        Number of digits to display to the right of the decimal in scientific notation (default 3).
    E_THRESHOLD :
        Power of 10 at which the float flips to scientific notation 10**+/- (default 6).
    E_MIN : int or None, optional
        lower limit before going to scientific notation (default None).
    E_MAX : int or None, optional
        Upper limit before going to scientific notation (default None).
    P_THRESHOLD : int or None, optional
        Precision threshold for area in between - so small values don't display as zero (default None).
    NUMBER_SEPARATOR : bool
        Flag for separating thousands in floats and ints (default False).
    NUMBER_SEPARATOR_CHAR : DisplayNumberSeparator
        Character for separating `,`, `.`, or `_` (default DisplayNumberSeparator.Comma).
    GB_PREFIX : str
         Prefix for column names to indicate that they are groupby keys (default "*").

    Methods
    -------
    save_config
    load_config
    reset_config
    e_min
    e_max
    p_threshold
    no_colors

    See also
    --------
    DisplayTable : builds result tables with display options.

    DisplayColorMode : enumerates supported color modes.

    get_terminal_size : calculates console height and width.

    Examples
    --------
    `CONSOLE_X_HTML` sets the number of characters for buffer width on the HTML display.
    Truncated characters are replaced by ellipsis.

    >>> from riptable.Utils.display_options import DisplayOptions
    >>> ds = rt.Dataset({'A':[0,6,9], 'B' : [1.2,3.1,9.6], 'C': [-1.6,2.7,4.6], 'D' : [2.4,6.2,19.2]})
    >>> ds
    #   A      B       C       D
    -   -   ----   -----   -----
    0   0   1.20   -1.60    2.40
    1   6   3.10    2.70    6.20
    2   9   9.60    4.60   19.20
    >>> DisplayOptions.CONSOLE_X_HTML = 25
    >>> ds
    #	A	B	...	D
    0	0	1.20	...	2.40
    1	6	3.10	...	6.20
    2	9	9.60	...	19.20
    """
    # class related options
    _CONFIG_LOADED = False  # default config file was found and loaded
    _AUTO_SAVE = False  # if true, config file will be saved to default path each time an option changes
    _USERNAME = None  # to distinguish between windows/linux environments
    _RESET_OPTIONS = False  # a flag so the next session will replace custom user options with the default options

    # screen/environment
    # Todo alz 20191125 - revisit the implementation, couldn't find usages and doesn't behave with console buffer
    CONSOLE_X_BUFFER = 30  # overall x buffer for console display
    CONSOLE_X_HTML = 340  # default "console width" for html display
    CONSOLE_X = 150  # default console width (also calculated by terminalsize.py) TODO: remove
    CONSOLE_Y = 25  # default console height (also calculated by terminalsize.py) TODO: remove
    HTML_DISPLAY = True  # force html display (TODO: remove)
    X_PADDING = 4  # character buffer for each column in console
    Y_PADDING = 3  # character buffer for each row in console

    # dataset/multiset
    ROW_ALL = False  # force all rows to display
    COL_ALL = False  # force all columns to display
    COL_MIN = 1  # min columns to display
    COL_MAX = 50  # max columns to display
    COL_T = 8  # number of transposed rows to display (which appear as columns)
    HEAD_ROWS = 15  # for dataset head
    TAIL_ROWS = 15  # for dataset tail
    MAX_ROWS = 30  # max rows to display
    NO_STYLES = False  # toggle for colors in the ipython console (sometimes hard to see with light background)
    COLOR_MODE = None  # set a color mode
    # NROWS_TRANSPOSE = 0 #
    # NCOLS_TRANSPOSE = 0 # if > 0, a specific number of
    # BORDER     = True # add a border beneath header labels
    # toggle so completion results show in alphanumeric key, attribute, then method ordering for Dataset, Multiset,
    # and Struct at any nested level
    CUSTOM_COMPLETION: bool = False

    # formatting for datasets/mutilsets/etc
    MAX_HEADER_WIDTH = 15  # maximum for header strings in dataset/multiset
    MAX_FOOTER_WIDTH = 15  # maximum for footer strings in dataset/multiset
    MAX_STRING_WIDTH = 15  # maximum for ALL strings

    # formatting for floating point and integer
    PRECISION = 2  # number of digits to the right of the decimal
    E_PRECISION = 3  # number of digits to display to the right of the decimal (sci notation)
    E_THRESHOLD = 6  # power of 10 at which the float flips to scientific notation 10**+/-

    E_MIN = None  # lower limit before going to scientific notation
    E_MAX = None  # upper limit before going to scientific notation
    P_THRESHOLD = None  # precision threshold for area in between - so small values don't display as zero

    NUMBER_SEPARATOR = False  # flag for separating thousands in floats and ints
    NUMBER_SEPARATOR_CHAR = DisplayNumberSeparator.Comma  # character for separating , . or _

    # misc
    GB_PREFIX = "*"  # prefix for column names to indicate that they are groupby keys

    # TODO: split the json config loader to separate files so that new display formatting
    # can be added more easily for future data types

    # min/max values for each display option
    # TODO: clean up and test DisplayOptions.__setitem__ to make sure these are followed
    _BOUNDS = {
        "HEAD_ROWS": (4, 500),
        "TAIL_ROWS": (4, 500),
        "MAX_ROWS": (4, 500),
        "COL_MIN": (1, 1000),
        "COL_MAX": (1, 51),
        "COL_T": (1, 200),
        "MAX_HEADER_WIDTH": (0, 100),
        "MAX_FOOTER_WIDTH": (0, 100),
        "MAX_STRING_WIDTH": (0, 100),
        "PRECISION": (1, 14),
        "E_PRECISION": (0.00000000001, 6),
        "E_THRESHOLD": (1, 20),
        "X_PADDING": (1, 25),
        "Y_PADDING": (1, 15),
        "CONSOLE_X": (80, 500),
        "CONSOLE_Y": (25, 150),
    }

    # test flags for display styling
    # see rt_display
    _PAINT_SIGNS = False
    _PAINT_ZEROS = False
    _PAINT_MIN = False
    _PAINT_MAX = False
    _BAR_GRAPH = False
    _TRANSPOSE = False
    _HEAT_MAP = False

    # test flag for including a footer in dataset/multiset display
    # TODO: more testing when a new Table class is created
    # header/footer/left/right
    _TEST_FOOTERS = False

    # test flag for reducing table printing to one pass over the data
    _TEST_ONE_PASS = False

    def __new__(cls):
        if cls._CONFIG_LOADED is False:
            DisplayOptions._get_username()
            was_loaded = DisplayOptions.load_config()
            if was_loaded is False:
                print("No display options found. Creating new display option config file...")
                try:
                    # may fail in windows
                    DisplayOptions.save_config()
                except Exception:
                    pass
            # reset code
            elif was_loaded == -1:
                DisplayOptions.save_config(force_overwrite=True)
        return cls

    def __init__(self):
        if self._CONFIG_LOADED is False:
            DisplayOptions._get_username()
            was_loaded = DisplayOptions.load_config()
            if was_loaded is False:
                print("No display options found. Creating new display option config file...")
                DisplayOptions.save_config()
            # reset code
            elif was_loaded == -1:
                DisplayOptions.save_config(force_overwrite=True)

    # when a new item is replaced or added ---------------------------
    def __setattr__(self, name, value):
        if hasattr(self, name):
            if name[0] == "_":
                setattr(self, name, value)

            else:
                bmin, bmax = self._BOUNDS[name]
                # old, for restricting property values
                # if value > bmax:
                #    pass
                # elif value < bmin:
                #    pass
                # else:
                #    pass

                # reset precision properties
                if name in ["PRECISION", "E_PRECISION", "E_THRESHOLD"]:
                    self.E_MIN = None
                    self.E_MAX = None
                    self.P_THRESHOLD = None

                setattr(self, name, value)

            if self._AUTO_SAVE:
                self.save_config()
        else:
            raise NameError(f"DisplayOptions has no attribute '{name}'.")

    @staticmethod
    def _get_username():
        if "USERNAME" in os.environ:
            DisplayOptions._USERNAME = os.environ["USERNAME"]
        elif "LOGNAME" in os.environ:
            DisplayOptions._USERNAME = os.environ["LOGNAME"]
        else:
            DisplayOptions._USERNAME = "riptable"

    @staticmethod
    def _get_default_path():
        # maybe store this to a class global?
        app_name = "riptable"
        # app_author = DisplayOptions._USERNAME
        config_dir = user_config_dir(app_name)

        return config_dir

    @staticmethod
    def save_config(path: Optional[str] = None, name: Optional[str] = None, force_overwrite: bool = False) -> bool:
        """
        Save display options at the default config file path if path and name are not supplied.
        Otherwise save display options using path and name.
        If `force_overwrite` is True, then silently overwrite any previous configs at that file path, otherwise prompt
        for user input before overwrite.

        Parameters
        ----------
        path : str, optional
            Path to display config
        name : str, optional
            Name of display config
        force_overwrite : bool
            True to overwrite if file already exists, otherwise prompt user whether to overwrite the file.

        Returns
        -------
        result : bool
            True if config was saved, otherwise False.
        """
        # get default save location, or accept a custom one
        # use this for now
        if path is None:
            path = DisplayOptions._get_default_path()
            if not os.path.exists(path):
                print("Default directory", path, "doesn't exist.")
                print("Creating directory...")
                # make an riptable directory if it doesn't exist
                try:
                    os.makedirs(path)
                except:
                    print("Unable to create directory", path)
                    return False
                else:
                    print("Success.")
            # add condition for generating default path for linux users
        if name is None:
            # fixes bug with autoreload
            if DisplayOptions._USERNAME is None:
                DisplayOptions._get_username()
            name = DisplayOptions._USERNAME + "_display_options.json"
        full_path = path + os.path.sep + name

        if os.path.exists(full_path) is True:
            if DisplayOptions._AUTO_SAVE is False:
                # if auto save is turned off, double check with user to make sure overwrite is okay
                if force_overwrite is False:
                    overwrite_prompt = "File exists at " + full_path + ". Overwrite? (y/n) "
                    overwrite = input(overwrite_prompt)
                    if overwrite != "y" and overwrite != "Y":
                        return False

        save_dict = {}
        for var in dir(DisplayOptions):
            # don't store functions or python default variables
            if not (callable(getattr(DisplayOptions, var))) and not var.startswith("__"):
                save_dict[var] = DisplayOptions.__getattribute__(DisplayOptions, var)
        try:
            with open(full_path, "w") as save_file:
                dump(save_dict, save_file)
        except:
            print("Error saving display options to", full_path)
            return False
        else:
            if force_overwrite is False:
                print("Saved display options as", name)
            return True

    @staticmethod
    def load_config(path: Optional[str] = None, name: Optional[str] = None) -> Union[bool, int]:
        """
        Load display config file from the default location if path and name are not supplied.
        Otherwise load display config settings using path and name.
        Return bool if applied correctly, otherwise return -1 if resetting display options.

        Parameters
        ----------
        path : str, optional
            Path to display config file
        name : str, optional
            Name of display config file

        Returns
        -------
        result : bool or int
            True if config was loaded correctly, otherwise False.
            -1 if a new default config will be saved after a reset
        """
        if path is None:
            try:
                # fails on some windows configs
                path = DisplayOptions._get_default_path()
            except Exception:
                return False
        if name is None:
            DisplayOptions._get_username()
            name = DisplayOptions._USERNAME + "_display_options.json"
        full_path = path + os.path.sep + name

        if not os.path.exists(full_path):
            print("No configuration file found at", full_path)
            return False

        try:
            with open(full_path, "r") as config_file:
                data = load(config_file)
        except:
            print("Error loading display options from", full_path)
            return False
        else:
            # a new default config file will be saved after a reset
            reset = data.get("_RESET_OPTIONS", False)
            # Todo alz 20191125 - use tuple rather than returning a value of different types
            # PEP 20 - Explicit is better than implicit and readability counts.
            if reset:
                print("Resetting display options.")
                return -1
            # print("Loaded display options.")
            for var, value in data.items():
                setattr(DisplayOptions, var, value)
            DisplayOptions._CONFIG_LOADED = True
            return True

    @staticmethod
    def reset_config(path: Optional[str] = None, name: Optional[str] = None) -> None:
        """
        Reapply display config file from default location if path and name are not supplied.
        Otherwise override display config settings using path and name.

        Parameters
        ----------
        path : str, optional
            Path to display config file
        name : str, optional
            Name of display config file

        Returns
        -------
        None
        """
        overwrite_prompt = "Are you sure you want to reset your display options? (y/n)"
        overwrite = input(overwrite_prompt)
        if overwrite != "y" and overwrite != "Y":
            return
        print("Options marked for reset. Start new session to complete reset.")
        DisplayOptions._RESET_OPTIONS = True
        # Todo alz 20191125 - implement use of path and name parameters
        DisplayOptions.save_config(force_overwrite=True)

    @classmethod
    def e_min(cls) -> int:
        """
        Returns the lower limit integer before displaying in scientific notation.

        Returns
        -------
        e_min: int
            lower limit before going to scientific notation.
        """
        if cls.E_MIN is None:
            cls.E_MIN = 10 ** (-1 * TypeRegister.DisplayOptions.E_THRESHOLD)
        return cls.E_MIN

    @classmethod
    def e_max(cls) -> int:
        """
        Returns the upper limit integer before displaying in scientific notation.

        Returns
        -------
        e_max : int
            Upper limit before going to scientific notation
        """
        if cls.E_MAX is None:
            cls.E_MAX = 10**TypeRegister.DisplayOptions.E_THRESHOLD
        return cls.E_MAX

    @classmethod
    def p_threshold(cls) -> float:
        """
        Returns DisplayOption.P_THRESHOLD.
        Defaults to 10 ** (-1 * DisplayOptions.PRECISION) - 1e-5.

        Returns
        -------
        p_threshold : float
             The precision threshold for area in between - so small values don't display as zero
        """
        if cls.P_THRESHOLD is None:
            cls.P_THRESHOLD = 10 ** (-1 * TypeRegister.DisplayOptions.PRECISION) - 1e-5
        return cls.P_THRESHOLD

    @staticmethod
    def no_colors() -> None:
        """
        Turn off all non-default table styles.

        Returns
        -------
        str, optional
        """
        DisplayOptions._PAINT_SIGNS = False
        DisplayOptions._PAINT_ZEROS = False
        DisplayOptions._PAINT_MIN = False
        DisplayOptions._PAINT_MAX = False
        DisplayOptions._BAR_GRAPH = False


TypeRegister.DisplayOptions = DisplayOptions



import pytest

import riptable as rt


def get_doctest_dataset_data():
    return {
        "ds_simple_1": rt.Dataset({"A": [0, 1, 6, 7], "B": [1.2, 3.1, 9.6, 21]}),
        "ds_simple_2": rt.Dataset({"X": [0, 1, 6, 9], "C": [2.4, 6.2, 19.2, 53]}),
        "ds_complex_1": rt.Dataset({"A": [0, 6, 9, 11], "B": ["Q", "R", "S", "T"], "C": [2.4, 6.2, 19.2, 25.9]}),
        "ds_complex_2": rt.Dataset(
            {
                "A": [0, 1, 6, 10],
                "B": ["Q", "R", "R", "T"],
                "E": [1.5, 3.75, 11.2, 13.1],
            }
        ),
    }


@pytest.fixture(autouse=True)
def docstring_imports(doctest_namespace):
    import numpy

    doctest_namespace["np"] = doctest_namespace["numpy"] = numpy
    doctest_namespace["rt"] = doctest_namespace["riptable"] = rt

    # Optional dependencies.
    import pandas

    doctest_namespace["pd"] = doctest_namespace["pandas"] = pandas


@pytest.fixture(autouse=True)
def docstring_merge_datasets(doctest_namespace):
    doctest_namespace.update(get_doctest_dataset_data())


@pytest.fixture(scope="session", autouse=True)
def register_null_log_handler():
    """
    Session-level fixture that installs a top-level log handler (or formatter) at the DEBUG level (or anything higher than NOTSET).
    It formats the messages in the typical way then just throws away the result. The idea is to just
    force all logging code to run, even if guarded with something like ``if logger.isEnabledFor(logging.DEBUG):``
    so we exercise that code within the tests. This helps guard against bad logging code that's only
    discovered when debug-level logging is enabled.
    """

    import logging

    class RenderingNullHandler(logging.Handler):
        """
        A ``logging.Handler`` which forces messages to be formatted using the default formatter
        to verify formatting strings are correct, then discards the formatted messages.
        """

        def __init__(self):
            """
            Initialize the instance.
            """
            logging.Handler.__init__(self)

        def emit(self, record):
            """
            Emit a record. Just forces the message to be formatted using the default
            formatter, then discards the result.
            """
            _ = self.format(record)

    # Create an instance of the RenderingNullHandler then register it
    # at the top level for all levels higher than logging.NOTSET.
    rendering_handler = RenderingNullHandler()
    root_logger = logging.getLogger()
    root_logger.addHandler(rendering_handler)
    root_logger.setLevel(logging.NOTSET + 1)



# $Id: //Depot/Source/SFW/riptable/Python/core/riptable/tests/test_csv.py#8 $
import os
import sys
import unittest

import numpy as np

from riptable import rt_csv

testdir = os.getenv("TEST_DIR", os.path.join(os.path.dirname(__file__), "test_files"))


def sfloat(x):
    try:
        return float(x)
    except:
        return np.nan


def sint(x):
    try:
        return int(x)
    except:
        return sys.maxsize


def scfloat(x):
    if type(x) is str:
        x = x.replace(",", "")
    return sfloat(x)


def scint(x):
    if type(x) is str:
        x = x.replace(",", "")
    return sint(x)


class Utility_Test(unittest.TestCase):
    def test_00(self, fname=None):
        if fname is None:
            fname = os.path.join(testdir, "unicode_ex3.csv")

        with open(fname, encoding="utf-8") as infile:
            headers = [_tok.strip('"') for _tok in infile.readline().strip().split(",")]
        ident = lambda x: x
        conv0 = {_k: ident for _k in headers}
        # conv1 = {_k: str for _k in headers}
        dsets = {}
        for ii in range(4):
            dsets[f"v{ii}"] = rt_csv.load_csv_as_dataset(fname, headers, conv0, version=ii, skip_rows=1)
        for _k in dsets:
            self.assertTrue((dsets["v0"] == dsets[_k]).all(axis=None))
            self.assertEqual(dsets[_k].shape, (35, 5))

        for ii in range(4):
            dsets[f"v{ii}"] = rt_csv.load_csv_as_dataset(fname, headers, conv0, version=ii, skip_rows=6)
        for _k in dsets:
            self.assertTrue((dsets["v0"] == dsets[_k]).all(axis=None))
            self.assertEqual(dsets[_k].shape, (30, 5))

    # BUG in this test. calling pd.read_csv corrupts memory and changes tests that follow
    # def test_02(self):
    #    cnv = dict(KEY=str, Region=str, Total=scint, Hombre=scint, Mujer=scint)
    #    fname1 = os.path.join(testdir, 'unicode_ex1.csv')
    #    fname2 = os.path.join(testdir, 'unicode_ex2.csv')
    #    fname3 = os.path.join(testdir, 'unicode_ex3.csv')

    #    df1 = pd.read_csv(fname1, encoding='latin-1', converters=cnv)
    #    df2 = pd.read_csv(fname2, encoding='utf-8', converters=cnv)

    #    self.assertTrue((df1 == df2).all().all())
    #    df3 = pd.read_csv(fname3, encoding='utf-8', converters=cnv)
    #    ds1 = dataset_from_pandas_df(df1)
    #    ds2 = dataset_from_pandas_df(df2)
    #    ds3 = dataset_from_pandas_df(df3)
    #    self.assertEqual(ds1.shape, (33, 5))
    #    self.assertEqual(ds2.shape, (33, 5))
    #    self.assertEqual(ds3.shape, (35, 5))
    #    with open(fname3, encoding='utf-8') as infile:
    #        ds4 = rt_csv.load_csv_as_dataset(infile, None, cnv)
    #    self.assertTrue((ds4 == ds3).all(axis=None))
    #    with open(fname3, encoding='utf-8') as infile:
    #        ds5 = rt_csv.load_csv_as_dataset(infile, list(ds3.keys()), cnv, skip_rows=1)
    #    self.assertTrue((ds5 == ds3).all(axis=None))
    #    sum1 = ds3[ds3.Region > 'Zacatecas', :][['Total', 'Hombre', 'Mujer']].sum()
    #    sum2 = dataset_as_pandas_df(ds3[ds3.Region > 'Zacatecas', :][['Total', 'Hombre', 'Mujer']]).sum()
    #    # ^-- reorders my fields alphabetically
    #    self.assertEqual(sum1.tolist()[0], sum2.values.tolist())
    #    subdf1 = dataset_as_pandas_df(ds3[ds3.Region < 'Puebla', :])
    #    subdf2 = dataset_as_pandas_df(ds2[ds2.Region < 'Puebla', :])
    #    self.assertTrue((subdf1 == subdf2).all().all())
    #    self.assertTrue((ds3[ds3.Region < 'Puebla', :] == ds2[ds2.Region < 'Puebla', :]).all(axis=None))
    #    return (df1, df2, df3), (ds1, ds2, ds3)

    # def test_03(self):
    #    cnv = dict(KEY=str, Region=str, Total=scint, Hombre=scint, Mujer=scint)
    #    fname3 = os.path.join(testdir, 'unicode_ex3.csv')
    #    df3 = pd.read_csv(fname3, encoding='utf-8', converters=cnv)
    #    ds3 = dataset_from_pandas_df(df3)
    #    ds3['Cs'] = [_s.startswith('C') for _s in ds3.Region]
    #    ds3['l1'] = [_s[0] for _s in ds3.Region]
    #    gb3_sum = ds3.groupby(['l1', 'Cs']).sum()
    #    result = gb3_sum.asdict()
    #    fname4 = os.path.join(testdir, 'groupby1_ex3.pickle')
    #    with open(fname4, 'rb') as infile:
    #        comp = pickle.load(infile)

    #    # asdict() now returns groupby columns in addition to the rest of the columns
    #    # had to remove groupby columns from asdict() result for correct comparison
    #    gb_keys = gb3_sum.label_get_names()
    #    correct_headers = [ h for h in list(result) if h not in gb_keys ]
    #    self.assertEqual(list(comp), correct_headers)

    #    correct_values = [ v for k,v in result.items() if k not in gb_keys ]
    #    for _v1, _v2 in zip(comp.values(), correct_values):
    #        self.assertTrue((_v1 ==_v2).all())


if __name__ == "__main__":
    tester = unittest.main()



"""
rt: RIPTABLE
"""
__author__ = "SIG"
from ._version import __version__
from .rt_accum2 import Accum2
from .rt_accumtable import AccumTable, accum_cols, accum_ratio, accum_ratiop
from .rt_categorical import Categorical, CatZero, categorical_convert
from .rt_dataset import Dataset
from .rt_enum import (
    DATETIME_TYPES,
    DS_DISPLAY_TYPES,
    GB_FUNCTIONS,
    INVALID_DICT,
    MATH_OPERATION,
    REDUCE_FUNCTIONS,
    ROLLING_FUNCTIONS,
    SD_TYPES,
    SM_DTYPES,
    TIMEWINDOW_FUNCTIONS,
    DisplayArrayTypes,
    DisplayColumnColors,
    DisplayDetectModes,
    DisplayJustification,
    DisplayLength,
    NumpyCharTypes,
    TypeRegister,
)

# Cannot simply remove imports w/o resolving the TypeRegister initialization
# at the end of each module.  Note the call to validate_registry()!
# Fundamental?
from .rt_fastarray import FastArray, Ledger, Recycle, Threading
from .rt_groupby import GroupBy
from .rt_groupbynumba import GroupbyNumba
from .rt_groupbyops import GroupByOps
from .rt_ledger import MathLedger
from .rt_multiset import Multiset
from .rt_pdataset import PDataset
from .rt_pgroupby import PGroupBy
from .rt_sharedmemory import SharedMemory
from .rt_struct import Struct
from .rt_timezone import Calendar, TimeZone

TypeRegister.validate_registry()
del TypeRegister
SharedMemory.check_shared_memory_limit()

import numpy as np
import riptide_cpp as rc

# short hand class names
from numpy import nan
from numpy import ndarray as NPA

from .rt_bin import cut, qcut, quantile
from .rt_categorical import Categorical as Cat
from .rt_csv import load_csv_as_dataset
from .rt_datetime import (
    Date,
    DateBase,
    DateScalar,
    DateSpan,
    DateSpanScalar,
    DateTimeBase,
    DateTimeNano,
    DateTimeNanoScalar,
    TimeSpan,
    TimeSpanScalar,
    datestring_to_nano,
    datetimestring_to_nano,
    strptime_to_nano,
    timestring_to_nano,
)
from .rt_display import DisplayTable as Display
from .rt_enum import TypeRegister
from .rt_fastarray import FastArray as FA
from .rt_fastarraynumba import fill_backward, fill_forward
from .rt_grouping import Grouping, combine2groups, hstack_groupings, merge_cats
from .rt_hstack import hstack_any, stack_rows

# Not fundamental yet of first-rank?
from .rt_merge import merge, merge2, merge_asof, merge_lookup
from .rt_misc import (
    autocomplete,
    output_cache_flush,
    output_cache_none,
    output_cache_setsize,
    profile_func,
)
from .rt_mlutils import normalize_minmax, normalize_zscore

# Perhaps convenience?
from .rt_numpy import (
    abs,
    absolute,
    all,
    any,
    arange,
    argmax,
    argmin,
    argsort,
    asanyarray,
    asarray,
    assoc_copy,
    assoc_index,
    bincount,
    bitcount,
    bool_,
    bool_to_fancy,
    bytes_,
    cat2keys,
    ceil,
    combine2keys,
    combine_accum1_filter,
    combine_accum2_filter,
    combine_filter,
    concatenate,
    crc32c,
    crc64,
    cumprod,
    cumsum,
    diff,
    double,
    empty,
    empty_like,
    float32,
    float64,
    floor,
    full,
    get_common_dtype,
    get_dtype,
    groupby,
    groupbyhash,
    groupbylex,
    groupbypack,
    half,
    hstack,
    int0,
    int8,
    int16,
    int32,
    int64,
    interp,
    interp_extrap,
    isfinite,
    isinf,
    ismember,
    isnan,
    isnanorzero,
    isnotfinite,
    isnotinf,
    isnotnan,
    issorted,
    lexsort,
    log,
    log10,
    logical,
    longdouble,
    makeifirst,
    makeilast,
    makeinext,
    makeiprev,
    mask_and,
    mask_andi,
    mask_andnot,
    mask_andnoti,
    mask_or,
    mask_ori,
    mask_xor,
    mask_xori,
    max,
    maximum,
    mean,
    median,
    min,
    minimum,
    multikeyhash,
    nan_to_num,
    nan_to_zero,
    nanargmax,
    nanargmin,
    nanmax,
    nanmean,
    nanmedian,
    nanmin,
    nanpercentile,
    nanstd,
    nansum,
    nanvar,
    ones,
    ones_like,
    percentile,
    power,
    putmask,
    reindex_fast,
    repeat,
    reshape,
    round,
    searchsorted,
    single,
    sort,
    sortinplaceindirect,
    std,
    str_,
    sum,
    tile,
    transpose,
    trunc,
    uint0,
    uint8,
    uint16,
    uint32,
    uint64,
    unique,
    unique32,
    var,
    vstack,
    where,
    zeros,
    zeros_like,
)
from .rt_sds import (
    SDSVerboseOff,
    SDSVerboseOn,
    compress_dataset_internal,
    decompress_dataset_internal,
    load_sds,
    load_sds_mem,
    save_sds,
    save_struct,
    sds_concat,
    sds_flatten,
    sds_info,
    sds_tree,
)
from .rt_str import FAString

# Second-rank?
from .rt_timers import (
    GetNanoTime,
    GetTSC,
    tic,
    ticf,
    ticp,
    ticx,
    toc,
    tocf,
    tocp,
    tocx,
    tt,
    ttx,
    utcnow,
)
from .rt_utils import (
    alignmk,
    bytes_to_str,
    describe,
    findTrueWidth,
    get_default_value,
    h5io_to_struct,
    ischararray,
    islogical,
    load_h5,
    mbget,
    merge_prebinned,
    normalize_keys,
    str_to_bytes,
    to_str,
)
from .Utils.rt_display_nested import treedir
from .Utils.rt_testdata import TestData as td
from .Utils.rt_testdata import load_test_data

# matplotlib may not be installed.
try:
    from . import rt_matplotlib
except ModuleNotFoundError:
    pass

# if environment is IPython/Jupyter, force config options here
try:
    from IPython import get_ipython

    ipconfig = get_ipython()
except ModuleNotFoundError:
    ipconfig = None

if ipconfig is not None:
    # disable outputcache only for ipython by default
    output_cache_none()

    # turn greedy on for better autocompletions on riptable classes
    ipconfig.Completer.greedy = True

    # as of Dec 16, 2019 we let user control jedi
    # if jedi is False, then _def_complete will be called
    # ipconfig.Completer.use_jedi = False

    # as of Jan 15, 2020 we are turning on our autocompletion engine
    autocomplete()

    # If DisplayOptions.CUSTOM_COMPLETION is toggled then use custom attribute completion for riptable data objects.
    from .Utils import ipython_utils
